# -----------------------------------------------------------------------------
# Copyright (c) 2023, 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
5500 - Module for testing pools with asyncio
"""

import asyncio
import unittest

import oracledb
import test_env


@unittest.skipUnless(
    test_env.get_is_thin(), "asyncio not supported in thick mode"
)
class TestCase(test_env.BaseAsyncTestCase):
    require_connection = False

    async def __connect_and_drop(self, pool):
        async with pool.acquire() as conn:
            cursor = conn.cursor()
            await cursor.execute("select count(*) from TestNumbers")
            (count,) = await cursor.fetchone()
            self.assertEqual(count, 10)

    async def __connect_and_generate_error(self, pool):
        async with pool.acquire() as conn:
            cursor = conn.cursor()
            with self.assertRaisesFullCode("ORA-01476"):
                await cursor.execute("select 1 / 0 from dual")

    async def __verify_connection(
        self, connection, expected_user, expected_proxy_user=None
    ):
        cursor = connection.cursor()
        await cursor.execute(
            """
            select
                sys_context('userenv', 'session_user'),
                sys_context('userenv', 'proxy_user')
            from dual
            """
        )
        actual_user, actual_proxy_user = await cursor.fetchone()
        self.assertEqual(actual_user, expected_user.upper())
        self.assertEqual(
            actual_proxy_user,
            expected_proxy_user and expected_proxy_user.upper(),
        )
        await connection.close()

    async def __verify_create_arg(self, arg_name, arg_value, sql):
        args = {}
        args[arg_name] = arg_value
        pool = test_env.get_pool_async(**args)
        async with pool.acquire() as conn:
            cursor = conn.cursor()
            await cursor.execute(sql)
            (fetched_value,) = await cursor.fetchone()
            self.assertEqual(fetched_value, arg_value)
        await pool.close()

    async def test_5500(self):
        "5500 - test getting default pool parameters"
        pool = test_env.get_pool_async()
        try:
            self.assertEqual(pool.busy, 0)
            self.assertEqual(pool.dsn, test_env.get_connect_string())
            self.assertEqual(pool.getmode, oracledb.POOL_GETMODE_WAIT)
            self.assertTrue(pool.homogeneous)
            self.assertEqual(pool.increment, 1)
            self.assertEqual(pool.max, 2)
            self.assertEqual(pool.max_lifetime_session, 0)
            self.assertEqual(pool.min, 1)
            self.assertEqual(pool.ping_interval, 60)
            self.assertEqual(
                pool.stmtcachesize, oracledb.defaults.stmtcachesize
            )
            self.assertEqual(pool.thin, True)
            self.assertEqual(pool.timeout, 0)
            self.assertEqual(pool.username, test_env.get_main_user())
        finally:
            await pool.close(force=True)

    async def test_5501(self):
        "5501 - test setting pool attributes"
        pool = test_env.get_pool_async()
        test_values = [
            ((11, 2), "ping_interval", 30),
            ((11, 2), "stmtcachesize", 100),
            ((11, 2), "timeout", 10),
            ((12, 2), "getmode", oracledb.POOL_GETMODE_TIMEDWAIT),
            ((12, 1), "max_lifetime_session", 3),
        ]
        try:
            for version, attr_name, value in test_values:
                if test_env.get_client_version() >= version:
                    setattr(pool, attr_name, value)
                    self.assertEqual(getattr(pool, attr_name), value)
                    self.assertRaises(
                        TypeError, setattr, pool, attr_name, "invalid value"
                    )
        finally:
            await pool.close(force=True)

    async def test_5502(self):
        "5502 - connection rolls back before released back to the pool"
        pool = test_env.get_pool_async()
        conn = await pool.acquire()
        cursor = conn.cursor()
        await cursor.execute("truncate table TestTempTable")
        await cursor.execute("insert into TestTempTable (IntCol) values (1)")
        cursor.close()
        await pool.release(conn)
        pool = test_env.get_pool_async()
        conn = await pool.acquire()
        cursor = conn.cursor()
        await cursor.execute("select count(*) from TestTempTable")
        (count,) = await cursor.fetchone()
        self.assertEqual(count, 0)
        await conn.close()

    async def test_5503(self):
        "5503 - test session pool with multiple coroutines"
        pool = test_env.get_pool_async(min=5, max=20, increment=2)
        try:
            coroutines = [self.__connect_and_drop(pool) for i in range(20)]
            await asyncio.gather(*coroutines)
        finally:
            await pool.close(force=True)

    async def test_5504(self):
        "5504 - test session pool with multiple coroutines (with errors)"
        pool = test_env.get_pool_async(min=5, max=20, increment=2)
        try:
            coroutines = [
                self.__connect_and_generate_error(pool) for i in range(20)
            ]
            await asyncio.gather(*coroutines)
        finally:
            await pool.close(force=True)

    @unittest.skipIf(test_env.get_is_drcp(), "not supported with DRCP")
    async def test_5505(self):
        "5505 - test session pool with various types of purity"
        pool = test_env.get_pool_async(min=1, max=8, increment=1)

        # get connection and set the action
        action = "TEST_ACTION"
        conn = await pool.acquire()
        conn.action = action
        cursor = conn.cursor()
        await cursor.execute("select 1 from dual")
        cursor.close()
        await pool.release(conn)
        self.assertEqual(pool.opened, 1, "opened (1)")

        # verify that the connection still has the action set on it
        conn = await pool.acquire()
        cursor = conn.cursor()
        await cursor.execute(
            "select sys_context('userenv', 'action') from dual"
        )
        (result,) = await cursor.fetchone()
        self.assertEqual(result, action)
        cursor.close()
        await pool.release(conn)
        self.assertEqual(pool.opened, 1, "opened (2)")

        # get a new connection with new purity (should not have state)
        conn = await pool.acquire(purity=oracledb.PURITY_NEW)
        cursor = conn.cursor()
        await cursor.execute(
            "select sys_context('userenv', 'action') from dual"
        )
        (result,) = await cursor.fetchone()
        self.assertIsNone(result)
        cursor.close()
        await pool.release(conn)

    async def test_5506(self):
        "5506 - test dropping/closing a connection from the pool"
        pool = test_env.get_pool_async(min=1, max=5, increment=2)
        try:
            conns1 = [await pool.acquire() for _ in range(2)]
            conns2 = [
                await oracledb.connect_async(pool=pool) for _ in range(3)
            ]
            self.assertEqual(pool.busy, 5)
            self.assertEqual(pool.opened, 5)

            for conn in conns1:
                await pool.drop(conn)
            self.assertEqual(pool.busy, 3)
            self.assertEqual(pool.opened, 3)

            for conn in conns2:
                await conn.close()
            self.assertEqual(pool.busy, 0)
            self.assertEqual(pool.opened, 3)
        finally:
            await pool.close(force=True)

    async def test_5507(self):
        "5507 - test to ensure pure connections are being created correctly"
        pool = test_env.get_pool_async(min=1, max=2, increment=1)
        try:
            conn1 = await pool.acquire()
            conn2 = await pool.acquire()
            self.assertEqual(pool.opened, 2, "opened (1)")
            await pool.release(conn1)
            await pool.release(conn2)
            conn3 = await pool.acquire(purity=oracledb.PURITY_NEW)
            self.assertEqual(pool.opened, 2, "opened (2)")
            await pool.release(conn3)
        finally:
            await pool.close(force=True)

    async def test_5508(self):
        "5508 - test closing a pool normally with no connections checked out"
        pool = test_env.get_pool_async(min=1, max=8, increment=1)
        await pool.close()

    async def test_5509(self):
        "5509 - test closing a pool normally with connections checked out"
        pool = test_env.get_pool_async(min=1, max=8, increment=1)
        closed = False
        try:
            async with pool.acquire():
                with self.assertRaisesFullCode("DPY-1005"):
                    await pool.close()
                    closed = True
        finally:
            if not closed:
                await pool.close(force=True)

    async def test_5510(self):
        "5510 - test closing a pool forcibly"
        pool = test_env.get_pool_async(min=1, max=8, increment=1)
        async with pool.acquire():
            await pool.close(force=True)

    async def test_5511(self):
        "5511 - using the pool after it is closed raises an exception"
        pool = test_env.get_pool_async(min=1, max=8, increment=1)
        await pool.close()
        with self.assertRaisesFullCode("DPY-1002"):
            await pool.acquire()

    async def test_5512(self):
        "5512 - using the pool beyond max limit raises an error"
        pool = test_env.get_pool_async(min=1, max=2, increment=1)
        try:
            async with pool.acquire(), pool.acquire():
                pool.getmode = oracledb.POOL_GETMODE_NOWAIT
                with self.assertRaisesFullCode("DPY-4005"):
                    await pool.acquire()
        finally:
            await pool.close(force=True)

    async def test_5513(self):
        "5513 - callable session callback is executed for new connections"

        class Counter:
            num_calls = 0

            @classmethod
            async def session_callback(cls, conn, requested_tag):
                cls.num_calls += 1

        pool = test_env.get_pool_async(
            min=1,
            max=2,
            increment=1,
            session_callback=Counter.session_callback,
        )
        try:
            async with pool.acquire(), pool.acquire():
                pass
            async with pool.acquire(), pool.acquire():
                pass
            self.assertEqual(Counter.num_calls, 2)
        finally:
            await pool.close(force=True)

    @unittest.skipIf(test_env.get_is_drcp(), "not supported with DRCP")
    async def test_5514(self):
        "5514 - drop the pooled connection on receiving dead connection error"
        admin_conn = await test_env.get_admin_connection_async()
        pool = test_env.get_pool_async(min=2, max=2, increment=2)
        try:

            # acquire connections from the pool and kill all the sessions
            with admin_conn.cursor() as admin_cursor:
                for conn in [await pool.acquire() for i in range(2)]:
                    sid, serial = await self.get_sid_serial(conn)
                    sql = f"alter system kill session '{sid},{serial}'"
                    await admin_cursor.execute(sql)
                    await conn.close()
            self.assertEqual(pool.opened, 2)

            # when try to re-use the killed sessions error will be raised;
            # release all such connections
            for conn in [await pool.acquire() for i in range(2)]:
                with conn.cursor() as cursor:
                    with self.assertRaisesFullCode("DPY-4011"):
                        await cursor.execute("select user from dual")
                await conn.close()

            # if a free connection is available, it can be used; otherwise a
            # new connection will be created
            for conn in [await pool.acquire() for i in range(2)]:
                with conn.cursor() as cursor:
                    await cursor.execute("select user from dual")
                    (user,) = await cursor.fetchone()
                    self.assertEqual(user, test_env.get_main_user().upper())
                await conn.close()
            self.assertEqual(pool.opened, 2)
        finally:
            await pool.close(force=True)

    async def test_5515(self):
        "5515 - acquire a connection from an empty pool (min=0)"
        pool = test_env.get_pool_async(min=0, max=2, increment=2)
        try:
            async with pool.acquire() as conn:
                with conn.cursor() as cursor:
                    await cursor.execute("select user from dual")
                    (result,) = await cursor.fetchone()
                    self.assertEqual(result, test_env.get_main_user().upper())
        finally:
            await pool.close(force=True)

    async def test_5516(self):
        "5516 - get different object types from different connections"
        pool = test_env.get_pool_async(min=1, max=2, increment=1)
        try:
            async with pool.acquire() as conn:
                typ = await conn.gettype("UDT_SUBOBJECT")
                self.assertEqual(typ.name, "UDT_SUBOBJECT")
            async with pool.acquire() as conn:
                typ = await conn.gettype("UDT_OBJECTARRAY")
                self.assertEqual(typ.name, "UDT_OBJECTARRAY")
        finally:
            await pool.close(force=True)

    async def test_5517(self):
        "5517 - test creating a pool using a proxy user"
        user_str = f"{test_env.get_main_user()}[{test_env.get_proxy_user()}]"
        pool = test_env.get_pool_async(user=user_str)
        try:
            await self.__verify_connection(
                await pool.acquire(),
                test_env.get_proxy_user(),
                test_env.get_main_user(),
            )
        finally:
            await pool.close(force=True)

    @unittest.skipIf(test_env.get_is_drcp(), "not supported with DRCP")
    async def test_5518(self):
        "5518 - test acquiring conn from pool in LIFO order"
        pool = test_env.get_pool_async(min=5, max=10, increment=1)
        try:
            sql = "select sys_context('userenv', 'sid') from dual"
            conns = [await pool.acquire() for i in range(3)]
            sids = []
            for conn in conns:
                with conn.cursor() as cursor:
                    await cursor.execute(sql)
                    (sid,) = await cursor.fetchone()
                    sids.append(sid)
            await conns[1].close()
            await conns[2].close()
            await conns[0].close()

            async with pool.acquire() as conn:
                with conn.cursor() as cursor:
                    await cursor.execute(sql)
                    (sid,) = await cursor.fetchone()
                    self.assertEqual(sid, sids[0], "not LIFO")
        finally:
            await pool.close(force=True)

    async def test_5519(self):
        "5519 - verify that dynamic pool cannot have an increment of zero"
        pool = test_env.get_pool_async(min=1, max=3, increment=0)
        try:
            self.assertEqual(pool.increment, 1)
            async with pool.acquire(), pool.acquire():
                pass
        finally:
            await pool.close(force=True)

    async def test_5520(self):
        "5520 - verify that static pool can have an increment of zero"
        pool = test_env.get_pool_async(min=1, max=1, increment=0)
        try:
            self.assertEqual(pool.increment, 0)
            async with pool.acquire():
                pass
        finally:
            await pool.close(force=True)

    async def test_5521(self):
        "5521 - verify that connection with different cclass is reused"
        cclass = "cclass2431"
        pool = test_env.get_pool_async(min=1, max=1)
        # ignore the first acquire which, depending on the speed with which the
        # minimum connections are created, may create a connection that is
        # discarded; instead, use the second acquire which should remain in the
        # pool
        try:
            async with pool.acquire(cclass=cclass) as conn:
                pass
            async with pool.acquire(cclass=cclass) as conn:
                sid_serial = await self.get_sid_serial(conn)
            async with pool.acquire(cclass=cclass) as conn:
                next_sid_serial = await self.get_sid_serial(conn)
                self.assertEqual(next_sid_serial, sid_serial)
            self.assertEqual(pool.opened, 1)
        finally:
            await pool.close(force=True)

    async def test_5522(self):
        "5522 - test creating a pool invalid params"
        with self.assertRaisesFullCode("DPY-2027"):
            oracledb.create_pool_async(params="bad params")

    async def test_5523(self):
        "5523 - test releasing and dropping an invalid connection"
        pool = test_env.get_pool_async()
        try:
            with self.assertRaises(TypeError):
                await pool.release("invalid connection")
            with self.assertRaises(TypeError):
                await pool.drop("invalid connection")
        finally:
            await pool.close(force=True)

    async def test_5524(self):
        "5524 - test creating a pool with invalid pool_class"
        with self.assertRaisesFullCode("DPY-2026"):
            oracledb.create_pool_async(pool_class=int)

    async def test_5525(self):
        "5525 - test creating a pool with a subclassed connection type"

        class MyConnection(oracledb.AsyncConnection):
            pass

        pool = test_env.get_pool_async(connectiontype=MyConnection)
        async with pool.acquire() as conn:
            self.assertIsInstance(conn, MyConnection)

    async def test_5526(self):
        "5526 - test creating a pool with a subclassed pool type"

        class MyPool(oracledb.AsyncConnectionPool):
            pass

        pool = test_env.get_pool_async(pool_class=MyPool)
        try:
            self.assertIsInstance(pool, MyPool)
        finally:
            await pool.close(force=True)

    async def test_5527(self):
        "5527 - test connectiontype with an invalid connection class"
        with self.assertRaisesFullCode("DPY-2023"):
            test_env.get_pool_async(connectiontype=oracledb.Connection)
        with self.assertRaisesFullCode("DPY-2023"):
            test_env.get_pool_async(connectiontype=int)

    @unittest.skipIf(
        test_env.get_server_version() < (12, 2), "not supported on this server"
    )
    async def test_5528(self):
        "5528 - ensure that timed wait times out with appropriate exception"
        pool = test_env.get_pool_async(
            getmode=oracledb.POOL_GETMODE_TIMEDWAIT, min=0, wait_timeout=1
        )
        with self.assertRaisesFullCode("DPY-4005"):
            await pool.acquire()

    async def test_5529(self):
        "5529 - ensure call timeout is reset on connections returned by pool"
        pool = test_env.get_pool_async(ping_timeout=1000, ping_interval=0)
        async with pool.acquire() as conn:
            self.assertEqual(conn.call_timeout, 0)
        async with pool.acquire() as conn:
            self.assertEqual(conn.call_timeout, 0)

    async def test_5530(self):
        "5530 - test passing program when creating a pool"
        sql = (
            "select program from v$session "
            "where sid = sys_context('userenv', 'sid')"
        )
        await self.__verify_create_arg("program", "newprogram", sql)

    async def test_5531(self):
        "5531 - test passing machine when creating a pool"
        sql = (
            "select machine from v$session "
            "where sid = sys_context('userenv', 'sid')"
        )
        await self.__verify_create_arg("machine", "newmachine", sql)

    async def test_5532(self):
        "5532 - test passing terminal when creating a pool"
        sql = (
            "select terminal from v$session "
            "where sid = sys_context('userenv', 'sid')"
        )
        await self.__verify_create_arg("terminal", "newterminal", sql)

    async def test_5533(self):
        "5533 - test passing osuser when creating a pool"
        sql = (
            "select osuser from v$session "
            "where sid = sys_context('userenv', 'sid')"
        )
        await self.__verify_create_arg("osuser", "newosuser", sql)

    async def test_5534(self):
        "5534 - test passing driver_name when creating a pool"
        sql = (
            "select distinct client_driver from v$session_connect_info "
            "where sid = sys_context('userenv', 'sid')"
        )
        await self.__verify_create_arg("driver_name", "newdriver", sql)

    async def test_5535(self):
        "5535 - test register_parameter with pooled connection"
        sdu = 4096
        params = test_env.get_pool_params()
        protocol = "proto-test"
        orig_connect_string = test_env.get_connect_string()
        connect_string = f"{protocol}://{orig_connect_string}"

        def hook(passed_protocol, passed_protocol_arg, passed_params):
            self.assertEqual(passed_protocol, protocol)
            self.assertEqual(passed_protocol_arg, orig_connect_string)
            passed_params.parse_connect_string(passed_protocol_arg)
            passed_params.set(sdu=sdu)

        try:
            oracledb.register_protocol(protocol, hook)
            pool = oracledb.create_pool_async(
                dsn=connect_string, params=params
            )
            self.assertEqual(params.sdu, sdu)
            async with pool.acquire():
                pass
            await pool.close()
        finally:
            oracledb.register_protocol(protocol, None)

    async def test_5536(self):
        "5536 - test create_pool() with edition"
        edition = test_env.get_edition_name()
        pool = test_env.get_pool_async(edition=edition)
        async with pool.acquire() as conn:
            self.assertEqual(conn.edition, edition)
        await pool.close()

    async def test_5537(self):
        "5537 - test create_pool() and get_pool() with alias"
        alias = "pool_alias_5537"
        pool = test_env.get_pool_async(pool_alias=alias)
        self.assertIs(pool, oracledb.get_pool(alias))
        await pool.close()

    async def test_5538(self):
        "5538 - test create_pool() twice with the same alias"
        alias = "pool_alias_5538"
        pool = test_env.get_pool_async(pool_alias=alias)
        with self.assertRaisesFullCode("DPY-2055"):
            test_env.get_pool_async(pool_alias=alias)
        await pool.close()
        self.assertIsNone(oracledb.get_pool(alias))

    async def test_5539(self):
        "5539 - test acquire() with pool alias and stmtcachesize"
        alias = "pool_5539"
        stmtcachesize = 35
        test_env.get_pool_async(pool_alias=alias, stmtcachesize=stmtcachesize)
        async with oracledb.connect_async(pool_alias=alias) as conn:
            self.assertEqual(conn.stmtcachesize, stmtcachesize)
        await oracledb.get_pool(alias).close()

    async def test_5540(self):
        "5540 - test pool alias is case sensitive"
        alias = "pool_5540"
        test_env.get_pool_async(pool_alias=alias)
        self.assertIsNone(oracledb.get_pool(alias.upper()))
        with self.assertRaisesFullCode("DPY-2054"):
            await test_env.get_connection_async(pool_alias=alias.upper())
        await oracledb.get_pool(alias).close()

    async def test_5541(self):
        "5541 - test pool alias with invalid types"
        aliases = [5, set(), dict(), bytearray(1)]
        for alias in aliases:
            with self.subTest(alias=alias):
                with self.assertRaises(TypeError):
                    test_env.get_pool_async(pool_alias=alias)


if __name__ == "__main__":
    test_env.run_test_cases()



# -----------------------------------------------------------------------------
# Copyright (c) 2023, 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
5400 - Module for testing the cursor execute() method with asyncio
"""

import collections
import unittest

import oracledb
import test_env


@unittest.skipUnless(
    test_env.get_is_thin(), "asyncio not supported in thick mode"
)
class TestCase(test_env.BaseAsyncTestCase):
    async def test_5400(self):
        "5400 - test executing a statement without any arguments"
        result = await self.cursor.execute("begin null; end;")
        self.assertIsNone(result)

    async def test_5401(self):
        "5401 - test executing a None statement with bind variables"
        cursor = self.conn.cursor()
        with self.assertRaisesFullCode("DPY-2001"):
            await cursor.execute(None, x=5)

    async def test_5402(self):
        "5402 - test executing a statement with args and empty keyword args"
        simple_var = self.cursor.var(oracledb.NUMBER)
        args = [simple_var]
        kwargs = {}
        result = await self.cursor.execute(
            "begin :1 := 25; end;", args, **kwargs
        )
        self.assertIsNone(result)
        self.assertEqual(simple_var.getvalue(), 25)

    async def test_5403(self):
        "5403 - test executing a statement with keyword arguments"
        simple_var = self.cursor.var(oracledb.NUMBER)
        result = await self.cursor.execute(
            "begin :value := 5; end;", value=simple_var
        )
        self.assertIsNone(result)
        self.assertEqual(simple_var.getvalue(), 5)

    async def test_5404(self):
        "5404 - test executing a statement with a dictionary argument"
        simple_var = self.cursor.var(oracledb.NUMBER)
        dict_arg = dict(value=simple_var)
        result = await self.cursor.execute(
            "begin :value := 10; end;", dict_arg
        )
        self.assertIsNone(result)
        self.assertEqual(simple_var.getvalue(), 10)

    async def test_5405(self):
        "5405 - test executing a statement with both a dict and keyword args"
        simple_var = self.cursor.var(oracledb.NUMBER)
        dict_arg = dict(value=simple_var)
        with self.assertRaisesFullCode("DPY-2005"):
            await self.cursor.execute(
                "begin :value := 15; end;", dict_arg, value=simple_var
            )

    async def test_5406(self):
        "5406 - test executing a statement and then changing the array size"
        await self.cursor.execute("select IntCol from TestNumbers")
        self.cursor.arraysize = 5
        self.assertEqual(len(await self.cursor.fetchall()), 10)

    async def test_5407(self):
        "5407 - test that subsequent executes succeed after bad execute"
        sql = "begin raise_application_error(-20000, 'this); end;"
        with self.assertRaisesFullCode("DPY-2041"):
            await self.cursor.execute(sql)
        await self.cursor.execute("begin null; end;")

    async def test_5408(self):
        "5408 - test that subsequent fetches fail after bad execute"
        with self.assertRaisesFullCode("ORA-00904"):
            await self.cursor.execute("select y from dual")
        with self.assertRaisesFullCode("DPY-1003"):
            await self.cursor.fetchall()

    async def test_5409(self):
        "5409 - test executing a statement with an incorrect named bind"
        sql = "select * from TestStrings where IntCol = :value"
        with self.assertRaisesFullCode("DPY-4008", "ORA-01036"):
            await self.cursor.execute(sql, value2=3)

    async def test_5410(self):
        "5410 - test executing a statement with named binds"
        await self.cursor.execute(
            """
            select *
            from TestNumbers
            where IntCol = :value1 and LongIntCol = :value2
            """,
            value1=1,
            value2=38,
        )
        self.assertEqual(len(await self.cursor.fetchall()), 1)

    async def test_5411(self):
        "5411 - test executing a statement with an incorrect positional bind"
        sql = """
                select *
                from TestNumbers
                where IntCol = :value and LongIntCol = :value2"""
        with self.assertRaisesFullCode("DPY-4009", "ORA-01008"):
            await self.cursor.execute(sql, [3])

    async def test_5412(self):
        "5412 - test executing a statement with positional binds"
        await self.cursor.execute(
            """
            select *
            from TestNumbers
            where IntCol = :value and LongIntCol = :value2
            """,
            [1, 38],
        )
        self.assertEqual(len(await self.cursor.fetchall()), 1)

    async def test_5413(self):
        "5413 - test executing a statement after rebinding a named bind"
        statement = "begin :value := :value2 + 5; end;"
        simple_var = self.cursor.var(oracledb.NUMBER)
        simple_var2 = self.cursor.var(oracledb.NUMBER)
        simple_var2.setvalue(0, 5)
        result = await self.cursor.execute(
            statement, value=simple_var, value2=simple_var2
        )
        self.assertIsNone(result)
        self.assertEqual(simple_var.getvalue(), 10)

        simple_var = self.cursor.var(oracledb.NATIVE_FLOAT)
        simple_var2 = self.cursor.var(oracledb.NATIVE_FLOAT)
        simple_var2.setvalue(0, 10)
        result = await self.cursor.execute(
            statement, value=simple_var, value2=simple_var2
        )
        self.assertIsNone(result)
        self.assertEqual(simple_var.getvalue(), 15)

    async def test_5414(self):
        "5414 - test executing a PL/SQL statement with duplicate binds"
        simple_var = self.cursor.var(oracledb.NUMBER)
        simple_var.setvalue(0, 5)
        result = await self.cursor.execute(
            """
            begin
                :value := :value + 5;
            end;
            """,
            value=simple_var,
        )
        self.assertIsNone(result)
        self.assertEqual(simple_var.getvalue(), 10)

    async def test_5415(self):
        "5415 - test executing a PL/SQL statement with duplicate binds"
        simple_var = self.cursor.var(oracledb.NUMBER)
        simple_var.setvalue(0, 5)
        await self.cursor.execute(
            "begin :value := :value + 5; end;", [simple_var]
        )
        self.assertEqual(simple_var.getvalue(), 10)

    async def test_5416(self):
        "5416 - test executing a statement with an incorrect number of binds"
        statement = "begin :value := :value2 + 5; end;"
        var = self.cursor.var(oracledb.NUMBER)
        var.setvalue(0, 5)
        with self.assertRaisesFullCode("DPY-4010", "ORA-01008"):
            await self.cursor.execute(statement)
        with self.assertRaisesFullCode("DPY-4010", "ORA-01008"):
            await self.cursor.execute(statement, value=var)
        with self.assertRaisesFullCode("DPY-4008", "ORA-01036"):
            await self.cursor.execute(
                statement, value=var, value2=var, value3=var
            )

    async def test_5417(self):
        "5417 - change in size on subsequent binds does not use optimised path"
        await self.cursor.execute("truncate table TestTempTable")
        data = [(1, "Test String #1"), (2, "ABC" * 100)]
        for row in data:
            await self.cursor.execute(
                """
                insert into TestTempTable (IntCol, StringCol1)
                values (:1, :2)
                """,
                row,
            )
        await self.conn.commit()
        await self.cursor.execute(
            "select IntCol, StringCol1 from TestTempTable"
        )
        self.assertEqual(await self.cursor.fetchall(), data)

    async def test_5418(self):
        "5418 - test that dml can use optimised path"
        data_to_insert = [(i + 1, f"Test String #{i + 1}") for i in range(3)]
        await self.cursor.execute("truncate table TestTempTable")
        for row in data_to_insert:
            with self.conn.cursor() as cursor:
                await cursor.execute(
                    """
                    insert into TestTempTable (IntCol, StringCol1)
                    values (:1, :2)
                    """,
                    row,
                )
        await self.conn.commit()
        await self.cursor.execute(
            "select IntCol, StringCol1 from TestTempTable order by IntCol"
        )
        self.assertEqual(await self.cursor.fetchall(), data_to_insert)

    async def test_5419(self):
        "5419 - test calling execute() with invalid parameters"
        sql = "insert into TestTempTable (IntCol, StringCol1) values (:1, :2)"
        with self.assertRaisesFullCode("DPY-2003"):
            await self.cursor.execute(sql, "These are not valid parameters")

    async def test_5420(self):
        "5420 - test calling execute() with mixed binds"
        await self.cursor.execute("truncate table TestTempTable")
        self.cursor.setinputsizes(None, None, str)
        data = dict(val1=1, val2="Test String 1")
        with self.assertRaisesFullCode("DPY-2006"):
            await self.cursor.execute(
                """
                insert into TestTempTable (IntCol, StringCol1)
                values (:1, :2)
                returning StringCol1 into :out_var
                """,
                data,
            )

    async def test_5421(self):
        "5421 - test binding by name with double quotes"
        data = {'"_value1"': 1, '"VaLue_2"': 2, '"3VALUE"': 3}
        await self.cursor.execute(
            'select :"_value1" + :"VaLue_2" + :"3VALUE" from dual',
            data,
        )
        (result,) = await self.cursor.fetchone()
        self.assertEqual(result, 6)

    async def test_5422(self):
        "5422 - test executing a statement with different input buffer sizes"
        sql = """
                insert into TestTempTable (IntCol, StringCol1, StringCol2)
                values (:int_col, :str_val1, :str_val2) returning IntCol
                into :ret_data"""
        values1 = {"int_col": 1, "str_val1": '{"a", "b"}', "str_val2": None}
        values2 = {"int_col": 2, "str_val1": None, "str_val2": '{"a", "b"}'}
        values3 = {"int_col": 3, "str_val1": '{"a"}', "str_val2": None}

        await self.cursor.execute("truncate table TestTempTable")
        ret_bind = self.cursor.var(oracledb.DB_TYPE_VARCHAR, arraysize=1)
        self.cursor.setinputsizes(ret_data=ret_bind)
        await self.cursor.execute(sql, values1)
        self.assertEqual(ret_bind.values, [["1"]])

        ret_bind = self.cursor.var(oracledb.DB_TYPE_VARCHAR, arraysize=1)
        self.cursor.setinputsizes(ret_data=ret_bind)
        await self.cursor.execute(sql, values2)
        self.assertEqual(ret_bind.values, [["2"]])

        ret_bind = self.cursor.var(oracledb.DB_TYPE_VARCHAR, arraysize=1)
        self.cursor.setinputsizes(ret_data=ret_bind)
        await self.cursor.execute(sql, values3)
        self.assertEqual(ret_bind.values, [["3"]])

    async def test_5423(self):
        "5423 - test using rowfactory"
        await self.cursor.execute("truncate table TestTempTable")
        await self.cursor.execute(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (1, 'Test 1')
            """
        )
        await self.conn.commit()
        await self.cursor.execute(
            "select IntCol, StringCol1 from TestTempTable"
        )
        column_names = [col[0] for col in self.cursor.description]

        def rowfactory(*row):
            return dict(zip(column_names, row))

        self.cursor.rowfactory = rowfactory
        self.assertEqual(self.cursor.rowfactory, rowfactory)
        self.assertEqual(
            await self.cursor.fetchall(),
            [{"INTCOL": 1, "STRINGCOL1": "Test 1"}],
        )

    async def test_5424(self):
        "5424 - test executing same query after setting rowfactory"
        await self.cursor.execute("truncate table TestTempTable")
        data = [(1, "Test 1"), (2, "Test 2")]
        await self.cursor.executemany(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (:1, :2)
            """,
            data,
        )
        await self.conn.commit()
        await self.cursor.execute(
            "select IntCol, StringCol1 from TestTempTable"
        )
        column_names = [col[0] for col in self.cursor.description]
        self.cursor.rowfactory = lambda *row: dict(zip(column_names, row))
        results1 = await self.cursor.fetchall()
        await self.cursor.execute(
            "select IntCol, StringCol1 from TestTempTable"
        )
        results2 = await self.cursor.fetchall()
        self.assertEqual(results1, results2)

    async def test_5425(self):
        "5425 - test executing different query after setting rowfactory"
        await self.cursor.execute("truncate table TestTempTable")
        data = [(1, "Test 1"), (2, "Test 2")]
        await self.cursor.executemany(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (:1, :2)
            """,
            data,
        )
        await self.conn.commit()
        await self.cursor.execute(
            "select IntCol, StringCol1 from TestTempTable"
        )
        column_names = [col[0] for col in self.cursor.description]
        self.cursor.rowfactory = lambda *row: dict(zip(column_names, row))
        await self.cursor.execute(
            """
            select IntCol, StringCol
            from TestSTrings
            where IntCol between 1 and 3 order by IntCol
            """
        )
        expected_data = [(1, "String 1"), (2, "String 2"), (3, "String 3")]
        self.assertEqual(await self.cursor.fetchall(), expected_data)

    async def test_5426(self):
        "5426 - test setting rowfactory on a REF cursor"
        with self.conn.cursor() as cursor:
            sql_function = "pkg_TestRefCursors.TestReturnCursor"
            ref_cursor = await cursor.callfunc(
                sql_function, oracledb.DB_TYPE_CURSOR, [2]
            )
            column_names = [col[0] for col in ref_cursor.description]
            ref_cursor.rowfactory = lambda *row: dict(zip(column_names, row))
            expected_value = [
                {"INTCOL": 1, "STRINGCOL": "String 1"},
                {"INTCOL": 2, "STRINGCOL": "String 2"},
            ]
            self.assertEqual(await ref_cursor.fetchall(), expected_value)

    async def test_5427(self):
        "5427 - test using a subclassed string as bind parameter keys"

        class my_str(str):
            pass

        await self.cursor.execute("truncate table TestTempTable")
        keys = {my_str("str_val"): oracledb.DB_TYPE_VARCHAR}
        self.cursor.setinputsizes(**keys)
        values = {
            my_str("int_val"): 5427,
            my_str("str_val"): "5427 - String Value",
        }
        await self.cursor.execute(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (:int_val, :str_val)
            """,
            values,
        )
        await self.cursor.execute(
            "select IntCol, StringCol1 from TestTempTable"
        )
        self.assertEqual(
            await self.cursor.fetchall(), [(5427, "5427 - String Value")]
        )

    async def test_5428(self):
        "5428 - test using a sequence of parameters other than a list or tuple"

        class MySeq(collections.abc.Sequence):
            def __init__(self, *data):
                self.data = data

            def __len__(self):
                return len(self.data)

            def __getitem__(self, index):
                return self.data[index]

        values_to_insert = [MySeq(1, "String 1"), MySeq(2, "String 2")]
        expected_data = [tuple(value) for value in values_to_insert]
        await self.cursor.execute("truncate table TestTempTable")
        await self.cursor.executemany(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (:int_val, :str_val)
            """,
            values_to_insert,
        )
        await self.cursor.execute(
            """
            select IntCol, StringCol1
            from TestTempTable
            order by IntCol
            """
        )
        self.assertEqual(await self.cursor.fetchall(), expected_data)

    async def test_5429(self):
        "5429 - test an output type handler with prefetch > arraysize"

        def type_handler(cursor, metadata):
            return cursor.var(metadata.type_code, arraysize=cursor.arraysize)

        self.cursor.arraysize = 2
        self.cursor.prefetchrows = 3
        self.cursor.outputtypehandler = type_handler
        await self.cursor.execute(
            "select level from dual connect by level <= 5"
        )
        self.assertEqual(
            await self.cursor.fetchall(), [(1,), (2,), (3,), (4,), (5,)]
        )

    async def test_5430(self):
        "5430 - test setinputsizes() but without binding"
        self.cursor.setinputsizes(None, int)
        sql = "select :1, : 2 from dual"

        with self.assertRaisesFullCode("ORA-01008", "DPY-4010"):
            await self.cursor.execute(sql, [])

    async def test_5431(self):
        "5431 - test getting FetchInfo attributes"
        type_obj = await self.conn.gettype("UDT_OBJECT")
        varchar_ratio, _ = await test_env.get_charset_ratios_async()
        test_values = [
            (
                "select IntCol from TestObjects",
                10,
                None,
                False,
                "INTCOL",
                False,
                9,
                0,
                oracledb.DB_TYPE_NUMBER,
                oracledb.DB_TYPE_NUMBER,
            ),
            (
                "select ObjectCol from TestObjects",
                None,
                None,
                False,
                "OBJECTCOL",
                True,
                None,
                None,
                type_obj,
                oracledb.DB_TYPE_OBJECT,
            ),
            (
                "select JsonVarchar from TestJsonCols",
                4000,
                4000 * varchar_ratio,
                True,
                "JSONVARCHAR",
                False,
                None,
                None,
                oracledb.DB_TYPE_VARCHAR,
                oracledb.DB_TYPE_VARCHAR,
            ),
            (
                "select FLOATCOL from TestNumbers",
                127,
                None,
                False,
                "FLOATCOL",
                False,
                126,
                -127,
                oracledb.DB_TYPE_NUMBER,
                oracledb.DB_TYPE_NUMBER,
            ),
        ]
        for (
            sql,
            display_size,
            internal_size,
            is_json,
            name,
            null_ok,
            precision,
            scale,
            typ,
            type_code,
        ) in test_values:
            await self.cursor.execute(sql)
            (fetch_info,) = self.cursor.description
            self.assertIsInstance(fetch_info, oracledb.FetchInfo)
            self.assertEqual(fetch_info.display_size, display_size)
            self.assertEqual(fetch_info.internal_size, internal_size)
            self.assertEqual(fetch_info.is_json, is_json)
            self.assertEqual(fetch_info.name, name)
            self.assertEqual(fetch_info.null_ok, null_ok)
            self.assertEqual(fetch_info.precision, precision)
            self.assertEqual(fetch_info.scale, scale)
            self.assertEqual(fetch_info.type, typ)
            self.assertEqual(fetch_info.type_code, type_code)

    async def test_5432(self):
        "5432 - test FetchInfo repr() and str()"
        await self.cursor.execute("select IntCol from TestObjects")
        (fetch_info,) = self.cursor.description
        self.assertEqual(
            str(fetch_info),
            "('INTCOL', <DbType DB_TYPE_NUMBER>, 10, None, 9, 0, False)",
        )
        self.assertEqual(
            repr(fetch_info),
            "('INTCOL', <DbType DB_TYPE_NUMBER>, 10, None, 9, 0, False)",
        )

    async def test_5433(self):
        "5433 - test slicing FetchInfo"
        await self.cursor.execute("select IntCol from TestObjects")
        (fetch_info,) = self.cursor.description
        self.assertEqual(fetch_info[1:3], (oracledb.DB_TYPE_NUMBER, 10))

    async def test_5434(self):
        "5434 - test async context manager"
        expected_value = test_env.get_main_user().upper()
        with self.conn.cursor() as cursor:
            await cursor.execute("select user from dual")
            self.assertEqual(await cursor.fetchone(), (expected_value,))
        async with self.conn.cursor() as cursor:
            await cursor.execute("select user from dual")
            self.assertEqual(await cursor.fetchone(), (expected_value,))

    async def test_5435(self):
        "5435 - test metadata requiring multiple packets"
        values = [f"Test value 5435 - {i}" for i in range(1, 301)]
        columns = ", ".join(f"'{v}'" for v in values)
        query = f"select {columns} from dual"
        await self.cursor.execute(query)
        row = await self.cursor.fetchone()
        self.assertEqual(row, tuple(values))

    async def test_5436(self):
        "5436 - test raising no_data_found in PL/SQL"
        with self.assertRaisesFullCode("ORA-01403"):
            await self.cursor.execute("begin raise no_data_found; end;")


if __name__ == "__main__":
    test_env.run_test_cases()



# -----------------------------------------------------------------------------
# Copyright (c) 2020, 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
4100 - Module for testing the methods for calling stored procedures and
functions (callproc() and callfunc())
"""

import oracledb
import test_env


class TestCase(test_env.BaseTestCase):
    def test_4100(self):
        "4100 - test executing a stored procedure"
        var = self.cursor.var(oracledb.NUMBER)
        results = self.cursor.callproc("proc_Test", ("hi", 5, var))
        self.assertEqual(results, ["hi", 10, 2.0])

    def test_4101(self):
        "4101 - test executing a stored procedure with all args keyword args"
        inout_value = self.cursor.var(oracledb.NUMBER)
        inout_value.setvalue(0, 5)
        out_value = self.cursor.var(oracledb.NUMBER)
        kwargs = dict(
            a_InOutValue=inout_value, a_InValue="hi", a_OutValue=out_value
        )
        results = self.cursor.callproc("proc_Test", [], kwargs)
        self.assertEqual(results, [])
        self.assertEqual(inout_value.getvalue(), 10)
        self.assertEqual(out_value.getvalue(), 2.0)

    def test_4102(self):
        "4102 - test executing a stored procedure with last arg as keyword arg"
        out_value = self.cursor.var(oracledb.NUMBER)
        kwargs = dict(a_OutValue=out_value)
        results = self.cursor.callproc("proc_Test", ("hi", 5), kwargs)
        self.assertEqual(results, ["hi", 10])
        self.assertEqual(out_value.getvalue(), 2.0)

    def test_4103(self):
        "4103 - test executing a stored procedure, repeated keyword arg"
        kwargs = dict(
            a_InValue="hi", a_OutValue=self.cursor.var(oracledb.NUMBER)
        )
        with self.assertRaisesFullCode("ORA-06550"):
            self.cursor.callproc("proc_Test", ("hi", 5), kwargs)

    def test_4104(self):
        "4104 - test executing a stored procedure without any arguments"
        results = self.cursor.callproc("proc_TestNoArgs")
        self.assertEqual(results, [])

    def test_4105(self):
        "4105 - test executing a stored function"
        results = self.cursor.callfunc("func_Test", oracledb.NUMBER, ("hi", 5))
        self.assertEqual(results, 7)

    def test_4106(self):
        "4106 - test executing a stored function without any arguments"
        results = self.cursor.callfunc("func_TestNoArgs", oracledb.NUMBER)
        self.assertEqual(results, 712)

    def test_4107(self):
        "4107 - test executing a stored function with wrong parameters"
        func_name = "func_Test"
        with self.assertRaisesFullCode("DPY-2007"):
            self.cursor.callfunc(oracledb.NUMBER, func_name, ("hi", 5))
        with self.assertRaisesFullCode("ORA-06550"):
            self.cursor.callfunc(func_name, oracledb.NUMBER, ("hi", 5, 7))
        with self.assertRaisesFullCode("DPY-2012"):
            self.cursor.callfunc(func_name, oracledb.NUMBER, "hi", 7)
        with self.assertRaisesFullCode("ORA-06502"):
            self.cursor.callfunc(func_name, oracledb.NUMBER, [5, "hi"])
        with self.assertRaisesFullCode("ORA-06550"):
            self.cursor.callfunc(func_name, oracledb.NUMBER)
        with self.assertRaisesFullCode("DPY-2012"):
            self.cursor.callfunc(func_name, oracledb.NUMBER, 5)

    def test_4108(self):
        "4108 - test to verify keywordParameters is deprecated"
        out_value = self.cursor.var(oracledb.NUMBER)
        kwargs = dict(a_OutValue=out_value)
        with self.assertRaisesFullCode("DPY-2014"):
            self.cursor.callproc(
                "proc_Test", ("hi", 5), kwargs, keywordParameters=kwargs
            )
        extra_amount = self.cursor.var(oracledb.NUMBER)
        extra_amount.setvalue(0, 5)
        kwargs = dict(a_ExtraAmount=extra_amount, a_String="hi")
        with self.assertRaisesFullCode("DPY-2014"):
            self.cursor.callfunc(
                "func_Test",
                oracledb.NUMBER,
                [],
                kwargs,
                keywordParameters=kwargs,
            )

    def test_4109(self):
        "4109 - test error for keyword args with invalid type"
        kwargs = [5]
        with self.assertRaisesFullCode("DPY-2013"):
            self.cursor.callproc("proc_Test", [], kwargs)
        with self.assertRaisesFullCode("DPY-2013"):
            self.cursor.callfunc("func_Test", oracledb.NUMBER, [], kwargs)

    def test_4110(self):
        "4110 - test to verify that deprecated keywordParameters works"
        extra_amount = self.cursor.var(oracledb.DB_TYPE_NUMBER)
        extra_amount.setvalue(0, 5)
        kwargs = dict(a_ExtraAmount=extra_amount, a_String="hi")
        results = self.cursor.callfunc(
            "func_Test", oracledb.DB_TYPE_NUMBER, keywordParameters=kwargs
        )
        self.assertEqual(results, 7)

        out_value = self.cursor.var(oracledb.DB_TYPE_NUMBER)
        kwargs = dict(a_OutValue=out_value)
        results = self.cursor.callproc(
            "proc_Test", ("hi", 5), keywordParameters=kwargs
        )
        self.assertEqual(results, ["hi", 10])
        self.assertEqual(out_value.getvalue(), 2.0)

    def test_4111(self):
        "4111 - test callproc with setinputsizes"
        out_value = self.cursor.var(oracledb.DB_TYPE_BOOLEAN)
        self.cursor.setinputsizes(
            oracledb.DB_TYPE_VARCHAR, oracledb.DB_TYPE_NUMBER, out_value
        )
        results = self.cursor.callproc("proc_Test2", ("hi", 5, out_value))
        self.assertEqual(results, ["hi", 10, True])
        self.assertTrue(out_value.getvalue())

    def test_4112(self):
        "4112 - test callfunc with setinputsizes"
        self.cursor.setinputsizes(
            oracledb.DB_TYPE_NUMBER,
            oracledb.DB_TYPE_VARCHAR,
            oracledb.DB_TYPE_NUMBER,
            oracledb.DB_TYPE_BOOLEAN,
        )
        results = self.cursor.callfunc(
            "func_Test2", oracledb.NUMBER, ("hi", 5, True)
        )
        self.assertEqual(results, 7)

    def test_4113(self):
        "4113 - test callproc with setinputsizes with kwargs"
        out_value = self.cursor.var(oracledb.DB_TYPE_BOOLEAN)
        self.cursor.setinputsizes(
            oracledb.DB_TYPE_VARCHAR, oracledb.DB_TYPE_NUMBER, out_value
        )
        kwargs = dict(a_OutValue=out_value)
        results = self.cursor.callproc("proc_Test2", ("hi", 5), kwargs)
        self.assertEqual(results, ["hi", 10])
        self.assertTrue(out_value.getvalue())

        out_value = self.cursor.var(oracledb.DB_TYPE_BOOLEAN)
        self.cursor.setinputsizes(
            oracledb.DB_TYPE_VARCHAR, oracledb.DB_TYPE_NUMBER, out_value
        )
        kwargs = dict(a_InValue="hi", a_InOutValue=5, a_OutValue=out_value)
        results = self.cursor.callproc("proc_Test2", [], kwargs)
        self.assertEqual(results, [])
        self.assertTrue(out_value.getvalue())

        self.cursor.setinputsizes(
            oracledb.DB_TYPE_VARCHAR,
            oracledb.DB_TYPE_NUMBER,
            oracledb.DB_TYPE_BOOLEAN,
        )
        kwargs = dict(a_InValue="hi", a_InOutValue=5, a_OutValue=out_value)
        results = self.cursor.callproc("proc_Test2", [], kwargs)
        self.assertEqual(results, [])
        self.assertTrue(out_value.getvalue())

    def test_4114(self):
        "4114 - test callproc with setinputsizes with kwargs in mixed order"
        out_value = self.cursor.var(oracledb.DB_TYPE_BOOLEAN)
        self.cursor.setinputsizes(
            oracledb.DB_TYPE_VARCHAR, oracledb.DB_TYPE_NUMBER, out_value
        )
        kwargs = dict(a_OutValue=out_value, a_InValue="hi", a_InOutValue=5)
        with self.assertRaisesFullCode("ORA-06550"):
            results = self.cursor.callproc(
                "proc_Test2", keyword_parameters=kwargs
            )
            self.assertEqual(results, [])
            self.assertTrue(out_value.getvalue())

        self.cursor.setinputsizes(
            oracledb.DB_TYPE_VARCHAR,
            oracledb.DB_TYPE_NUMBER,
            oracledb.DB_TYPE_BOOLEAN,
        )
        with self.assertRaisesFullCode("ORA-06550"):
            self.cursor.callproc("proc_Test2", keyword_parameters=kwargs)

    def test_4115(self):
        "4115 - test callfunc with setinputsizes with kwargs"
        extra_amount = self.cursor.var(oracledb.DB_TYPE_NUMBER)
        extra_amount.setvalue(0, 5)
        test_values = [
            (["hi"], dict(a_ExtraAmount=extra_amount, a_Boolean=True)),
            (
                [],
                dict(
                    a_String="hi", a_ExtraAmount=extra_amount, a_Boolean=True
                ),
            ),
        ]
        for args, kwargs in test_values:
            self.cursor.setinputsizes(
                oracledb.DB_TYPE_NUMBER,
                oracledb.DB_TYPE_VARCHAR,
                oracledb.DB_TYPE_NUMBER,
                oracledb.DB_TYPE_BOOLEAN,
            )
            results = self.cursor.callfunc(
                "func_Test2", oracledb.DB_TYPE_NUMBER, args, kwargs
            )
            self.assertEqual(results, 7)

    def test_4116(self):
        "4116 - test callproc with setinputsizes with extra arguments"
        out_value = self.cursor.var(oracledb.DB_TYPE_BOOLEAN)
        test_values = [
            (("hi", 5, out_value), None),
            (("hi",), dict(a_InOutValue=5, a_OutValue=out_value)),
            ([], dict(a_InValue="hi", a_InOutValue=5, a_OutValue=out_value)),
        ]
        for args, kwargs in test_values:
            self.cursor.setinputsizes(
                oracledb.DB_TYPE_VARCHAR,
                oracledb.NUMBER,
                out_value,
                oracledb.DB_TYPE_VARCHAR,  # extra argument
            )
            with self.assertRaisesFullCode("ORA-01036", "DPY-4009"):
                self.cursor.callproc("proc_Test2", args, kwargs)

    def test_4117(self):
        "4117 - test callfunc with setinputsizes with extra arguments"
        extra_amount = self.cursor.var(oracledb.DB_TYPE_NUMBER)
        extra_amount.setvalue(0, 5)
        test_values = [
            (["hi", extra_amount], None),
            (["hi"], dict(a_ExtraAmount=extra_amount)),
            ([], dict(a_ExtraAmount=extra_amount, a_String="hi")),
        ]
        for args, kwargs in test_values:
            self.cursor.setinputsizes(
                oracledb.DB_TYPE_NUMBER,
                oracledb.DB_TYPE_VARCHAR,
                oracledb.DB_TYPE_NUMBER,
                oracledb.DB_TYPE_BOOLEAN,
                oracledb.DB_TYPE_VARCHAR,  # extra argument
            )
            with self.assertRaisesFullCode("ORA-01036", "DPY-4009"):
                self.cursor.callfunc(
                    "func_Test2", oracledb.DB_TYPE_NUMBER, args, kwargs
                )

    def test_4118(self):
        "4118 - test callproc with setinputsizes with too few parameters"
        out_value = self.cursor.var(oracledb.DB_TYPE_BOOLEAN)

        # setinputsizes for 2 args (missed 1 args)
        self.cursor.setinputsizes(
            oracledb.DB_TYPE_VARCHAR, oracledb.DB_TYPE_NUMBER
        )
        results = self.cursor.callproc("proc_Test2", ("hi", 5, out_value))
        self.assertEqual(results, ["hi", 10, out_value.getvalue()])
        self.assertTrue(out_value.getvalue())

        # setinputsizes for 2 args (missed 1 kwargs)
        self.cursor.setinputsizes(
            oracledb.DB_TYPE_VARCHAR, oracledb.DB_TYPE_NUMBER
        )
        kwargs = dict(a_OutValue=out_value)
        results = self.cursor.callproc("proc_Test2", ("hi", 5), kwargs)
        self.assertEqual(results, ["hi", 10])
        self.assertTrue(out_value.getvalue())

        # setinputsizes for 1 args (missed 2 args)
        self.cursor.setinputsizes(oracledb.DB_TYPE_VARCHAR)
        results = self.cursor.callproc("proc_Test2", ("hi", 5, out_value))
        self.assertEqual(results, ["hi", 10, out_value.getvalue()])
        self.assertTrue(out_value.getvalue())

        # setinputsizes for 1 args (missed 1 args and 1 kwargs)
        self.cursor.setinputsizes(oracledb.DB_TYPE_VARCHAR)
        kwargs = dict(a_OutValue=out_value)
        results = self.cursor.callproc("proc_Test2", ("hi", 5), kwargs)
        self.assertEqual(results, ["hi", 10])
        self.assertTrue(out_value.getvalue())

        # setinputsizes for 2 kwargs (missed 1 kwargs)
        self.cursor.setinputsizes(
            oracledb.DB_TYPE_VARCHAR, oracledb.DB_TYPE_NUMBER
        )
        kwargs = dict(a_InValue="hi", a_InOutValue=5, a_OutValue=out_value)
        results = self.cursor.callproc("proc_Test2", [], kwargs)
        self.assertEqual(results, [])
        self.assertTrue(out_value.getvalue())

    def test_4119(self):
        """
        4119 - test callproc with setinputsizes with wrong order of parameters
        """
        # setinputsizes for 2 args (missed 1 kwargs)
        out_value = self.cursor.var(oracledb.DB_TYPE_BOOLEAN)
        self.cursor.setinputsizes(bool, oracledb.DB_TYPE_VARCHAR)
        kwargs = dict(a_OutValue=out_value)
        with self.assertRaisesFullCode("ORA-06550"):
            self.cursor.callproc("proc_Test2", ["hi", 5], kwargs)

        # setinputsizes for 2 kwargs (missed 1 kwargs)
        self.cursor.setinputsizes(bool, oracledb.DB_TYPE_VARCHAR)
        kwargs = dict(a_InValue="hi", a_InOutValue=5, a_OutValue=out_value)
        with self.assertRaisesFullCode("ORA-06550"):
            self.cursor.callproc("proc_Test2", [], kwargs)

    def test_4120(self):
        "4120 - test callfunc with setinputsizes with too few parameters"
        # setinputsizes for return_type and 1 kwargs (missed 2 kwargs)
        bool_var = self.cursor.var(oracledb.DB_TYPE_BOOLEAN)
        bool_var.setvalue(0, False)
        kwargs = dict(a_Boolean=bool_var, a_String="hi", a_ExtraAmount=3)
        self.cursor.setinputsizes(oracledb.NUMBER, oracledb.DB_TYPE_VARCHAR)
        results = self.cursor.callfunc(
            "func_Test2", oracledb.NUMBER, [], kwargs
        )
        self.assertEqual(results, -1)

        # setinputsizes for return_type (missed 3 kwargs)
        bool_var.setvalue(0, False)
        kwargs = dict(a_Boolean=bool_var, a_String="hi", a_ExtraAmount=1)
        self.cursor.setinputsizes(oracledb.NUMBER)
        results = self.cursor.callfunc(
            "func_Test2", oracledb.NUMBER, [], kwargs
        )
        self.assertEqual(results, 1)

        # setinputsizes for return_type (missed 3 args)
        bool_var.setvalue(0, True)
        self.cursor.setinputsizes(oracledb.NUMBER)
        results = self.cursor.callfunc(
            "func_Test2", oracledb.NUMBER, ["hi", 2, bool_var]
        )
        self.assertEqual(results, 4)

    def test_4121(self):
        """
        4121 - test callfunc with setinputsizes with wrong order of parameters
        """
        # setinputsizes for 2 args (missed 2 kwargs)
        bool_var = self.cursor.var(oracledb.DB_TYPE_BOOLEAN)
        bool_var.setvalue(0, True)
        self.cursor.setinputsizes(oracledb.NUMBER, oracledb.DB_TYPE_BOOLEAN)
        kwargs = dict(a_Boolean=bool_var)
        with self.assertRaisesFullCode("ORA-06550"):
            self.cursor.callfunc(
                "func_Test2", oracledb.NUMBER, ["hi", bool_var], kwargs
            )

    def test_4122(self):
        "4122 - test callfunc with setinputsizes without type for return_type"
        # setinputsizes for 1 args and 1 kwargs
        bool_var = self.cursor.var(oracledb.DB_TYPE_BOOLEAN)
        bool_var.setvalue(0, False)
        self.cursor.setinputsizes(oracledb.NUMBER, oracledb.DB_TYPE_BOOLEAN)
        kwargs = dict(a_Boolean=bool_var)
        with self.assertRaisesFullCode("ORA-06550"):
            self.cursor.callfunc(
                "func_Test2", oracledb.DB_TYPE_NUMBER, ["hi"], kwargs
            )

        # setinputsizes for 2 kwargs (missed 1 kwargs)
        bool_var.setvalue(0, False)
        kwargs = dict(a_Boolean=bool_var, a_String="hi", a_ExtraAmount=0)
        self.cursor.setinputsizes(
            oracledb.DB_TYPE_BOOLEAN, oracledb.DB_TYPE_VARCHAR
        )
        results = self.cursor.callfunc(
            "func_Test2", oracledb.DB_TYPE_NUMBER, [], kwargs
        )
        self.assertEqual(results, 2)

        # setinputsizes for 2 args and 1 kwargs
        bool_var.setvalue(0, False)
        self.cursor.setinputsizes(
            oracledb.DB_TYPE_BOOLEAN, oracledb.DB_TYPE_NUMBER
        )
        kwargs = dict(a_Boolean=bool_var)
        results = self.cursor.callfunc(
            "func_Test2", oracledb.DB_TYPE_NUMBER, ["Bye", 2], kwargs
        )
        self.assertEqual(results, 1)

        # setinputsizes for 2 args (missed 1 args)
        bool_var.setvalue(0, False)
        self.cursor.setinputsizes(
            oracledb.DB_TYPE_BOOLEAN, oracledb.DB_TYPE_NUMBER
        )
        kwargs = dict(a_Boolean=bool_var)
        results = self.cursor.callfunc(
            "func_Test2", oracledb.DB_TYPE_NUMBER, ["Light", -1, bool_var]
        )
        self.assertEqual(results, 6)

    def test_4123(self):
        "4123 - test executing a procedure with callfunc"
        with self.assertRaisesFullCode("ORA-06550"):
            self.cursor.callfunc(
                "proc_Test2", oracledb.NUMBER, ("hello", 3, False)
            )

    def test_4124(self):
        "4124 - test executing a function with callproc"
        with self.assertRaisesFullCode("ORA-06550"):
            self.cursor.callproc("func_Test2", ("hello", 5, True))

    def test_4125(self):
        "4125 - test calling a procedure with a string > 32767 characters"
        data = "4125" * 16000
        size_var = self.cursor.var(int)
        self.cursor.callproc("pkg_TestLobs.GetSize", [data, size_var])
        self.assertEqual(size_var.getvalue(), len(data))

    def test_4126(self):
        "4125 - test calling a procedure with raw data > 32767 bytes"
        data = b"4126" * 16250
        size_var = self.cursor.var(int)
        self.cursor.callproc("pkg_TestLobs.GetSize", [data, size_var])
        self.assertEqual(size_var.getvalue(), len(data))


if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# session_release.pyx
#
# Cython file defining the messages that are sent to the database and the
# responses that are received by the client for releasing a session (embedded
# in thin_impl.pyx).
#------------------------------------------------------------------------------

@cython.final
cdef class SessionReleaseMessage(Message):

    cdef:
        uint32_t release_mode

    cdef int _initialize_hook(self) except -1:
        """
        Perform initialization.
        """
        self.message_type = TNS_MSG_TYPE_ONEWAY_FN
        self.function_code = TNS_FUNC_SESSION_RELEASE

    cdef int _write_message(self, WriteBuffer buf) except -1:
        """
        Write the message for a DRCP session release.
        """
        self._write_function_code(buf)
        buf.write_uint8(0)                  # pointer (tag name)
        buf.write_uint8(0)                  # tag name length
        buf.write_ub4(self.release_mode)    # mode



#------------------------------------------------------------------------------
# Copyright (c) 2022, 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# dbobject.pyx
#
# Cython file defining the thin DbObjectType, DbObjectAttr and DbObject
# implementation classes (embedded in thin_impl.pyx).
#------------------------------------------------------------------------------

@cython.final
cdef class DbObjectPickleBuffer(GrowableBuffer):

    cdef int _read_raw_bytes_and_length(self, const char_type **ptr,
                                        ssize_t *num_bytes) except -1:
        """
        Helper function that processes the length (if needed) and then acquires
        the specified number of bytes from the buffer.
        """
        cdef uint32_t extended_num_bytes
        if num_bytes[0] == TNS_LONG_LENGTH_INDICATOR:
            self.read_uint32be(&extended_num_bytes)
            num_bytes[0] = <ssize_t> extended_num_bytes
        ptr[0] = self._get_raw(num_bytes[0])

    cdef int _write_raw_bytes_and_length(self, const char_type *ptr,
                                         ssize_t num_bytes) except -1:
        """
        Helper function that writes the length in the format required before
        writing the bytes.
        """
        self.write_length(num_bytes)
        self.write_raw(ptr, <uint32_t> num_bytes)

    cdef int get_is_atomic_null(self, bint is_collection,
                                bint* is_null) except -1:
        """
        Reads the next byte and checks to see if the value is atomically null.
        If not, the byte is returned to the buffer for further processing.
        """
        cdef uint8_t value
        self.read_ub1(&value)
        if value == TNS_OBJ_ATOMIC_NULL \
                or (is_collection and value == TNS_NULL_LENGTH_INDICATOR):
            is_null[0] = True
        else:
            is_null[0] = False
            self._pos -= 1

    cdef int read_header(self, uint8_t* flags, uint8_t *version) except -1:
        """
        Reads the header of the pickled data.
        """
        cdef:
            uint32_t prefix_seg_length
            uint8_t tmp
        self.read_ub1(flags)
        self.read_ub1(version)
        self.skip_length()
        if flags[0] & TNS_OBJ_IS_DEGENERATE:
            errors._raise_not_supported("DbObject stored in a LOB")
        if flags[0] & TNS_OBJ_NO_PREFIX_SEG:
            return 0
        self.read_length(&prefix_seg_length)
        self.skip_raw_bytes(prefix_seg_length)

    cdef int read_length(self, uint32_t *length) except -1:
        """
        Read the length from the buffer. This will be a single byte, unless the
        value meets or exceeds TNS_LONG_LENGTH_INDICATOR. In that case, the
        value is stored as a 4-byte integer.
        """
        cdef uint8_t short_length
        self.read_ub1(&short_length)
        if short_length == TNS_LONG_LENGTH_INDICATOR:
            self.read_uint32be(length)
        else:
            length[0] = short_length

    cdef object read_xmltype(self, BaseThinConnImpl conn_impl):
        """
        Reads an XML type from the buffer. This is similar to reading a
        database object but with specialized processing.
        """
        cdef:
            uint8_t image_flags, image_version
            BaseThinLobImpl lob_impl
            const char_type *ptr
            ssize_t bytes_left
            uint32_t xml_flag
            type cls
        self.read_header(&image_flags, &image_version)
        self.skip_raw_bytes(1)              # XML version
        self.read_uint32be(&xml_flag)
        if xml_flag & TNS_XML_TYPE_FLAG_SKIP_NEXT_4:
            self.skip_raw_bytes(4)
        bytes_left = self.bytes_left()
        ptr = self.read_raw_bytes(bytes_left)
        if xml_flag & TNS_XML_TYPE_STRING:
            return ptr[:bytes_left].decode()
        elif xml_flag & TNS_XML_TYPE_LOB:
            lob_impl = conn_impl._create_lob_impl(DB_TYPE_CLOB,
                                                  ptr[:bytes_left])
            cls = PY_TYPE_ASYNC_LOB \
                if conn_impl._protocol._transport._is_async \
                else PY_TYPE_LOB
            return cls._from_impl(lob_impl)
        errors._raise_err(errors.ERR_UNEXPECTED_XML_TYPE, flag=xml_flag)

    cdef int skip_length(self) except -1:
        """
        Skips the length instead of reading it from the buffer.
        """
        cdef uint8_t short_length
        self.read_ub1(&short_length)
        if short_length == TNS_LONG_LENGTH_INDICATOR:
            self.skip_raw_bytes(4)

    cdef int write_header(self, ThinDbObjectImpl obj_impl) except -1:
        """
        Writes the header of the pickled data. Since the size is unknown at
        this point, zero is written initially and the actual size is written
        later.
        """
        cdef ThinDbObjectTypeImpl typ_impl = obj_impl.type
        self.write_uint8(obj_impl.image_flags)
        self.write_uint8(obj_impl.image_version)
        self.write_uint8(TNS_LONG_LENGTH_INDICATOR)
        self.write_uint32be(0)
        if typ_impl.is_collection:
            self.write_uint8(1)             # length of prefix segment
            self.write_uint8(1)             # prefix segment contents

    cdef int write_length(self, ssize_t length) except -1:
        """
        Writes the length to the buffer.
        """
        if length <= TNS_OBJ_MAX_SHORT_LENGTH:
            self.write_uint8(<uint8_t> length)
        else:
            self.write_uint8(TNS_LONG_LENGTH_INDICATOR)
            self.write_uint32be(<uint32_t> length)


@cython.final
cdef class TDSBuffer(Buffer):
    pass


cdef class ThinDbObjectImpl(BaseDbObjectImpl):
    cdef:
        uint8_t image_flags, image_version
        bytes toid, oid, packed_data
        uint32_t num_elements
        dict unpacked_assoc_array
        list unpacked_assoc_keys
        dict unpacked_attrs
        list unpacked_array
        uint16_t flags

    cdef inline int _ensure_assoc_keys(self) except -1:
        """
        Ensure that the keys for the associative array have been calculated.
        PL/SQL associative arrays keep their keys in sorted order so this must
        be calculated when indices are required.
        """
        if self.unpacked_assoc_keys is None:
            self.unpacked_assoc_keys = list(sorted(self.unpacked_assoc_array))

    cdef inline int _ensure_unpacked(self) except -1:
        """
        Ensure that the data has been unpacked.
        """
        if self.packed_data is not None:
            self._unpack_data()

    cdef bytes _get_packed_data(self):
        """
        Returns the packed data for the object. This will either be the value
        retrieved from the database or generated packed data (for new objects
        and those that have had their data unpacked already).
        """
        cdef:
            ThinDbObjectTypeImpl typ_impl = self.type
            DbObjectPickleBuffer buf
            ssize_t size
        if self.packed_data is not None:
            return self.packed_data
        buf = DbObjectPickleBuffer.__new__(DbObjectPickleBuffer)
        buf._initialize()
        buf.write_header(self)
        self._pack_data(buf)
        size = buf._pos
        buf.skip_to(3)
        buf.write_uint32be(size)
        return buf._data[:size]

    cdef int _pack_data(self, DbObjectPickleBuffer buf) except -1:
        """
        Packs the data from the object into the buffer.
        """
        cdef:
            ThinDbObjectTypeImpl typ_impl = self.type
            ThinDbObjectAttrImpl attr
            int32_t index
            object value
        if typ_impl.is_collection:
            buf.write_uint8(typ_impl.collection_flags)
            if typ_impl.collection_type == TNS_OBJ_PLSQL_INDEX_TABLE:
                self._ensure_assoc_keys()
                buf.write_length(len(self.unpacked_assoc_keys))
                for index in self.unpacked_assoc_keys:
                    buf.write_uint32be(<uint32_t> index)
                    self._pack_value(buf, typ_impl.element_metadata,
                                     self.unpacked_assoc_array[index])
            else:
                buf.write_length(len(self.unpacked_array))
                for value in self.unpacked_array:
                    self._pack_value(buf, typ_impl.element_metadata, value)
        else:
            for attr in typ_impl.attrs:
                self._pack_value(buf, attr, self.unpacked_attrs[attr.name])

    cdef int _pack_value(self, DbObjectPickleBuffer buf,
                         OracleMetadata metadata, object value) except -1:
        """
        Packs a value into the buffer. At this point it is assumed that the
        value matches the correct type.
        """
        cdef:
            uint8_t ora_type_num = metadata.dbtype._ora_type_num
            ThinDbObjectImpl obj_impl
            BaseThinLobImpl lob_impl
            bytes temp_bytes
        if value is None:
            if metadata.objtype is not None \
                    and not metadata.objtype.is_collection:
                buf.write_uint8(TNS_OBJ_ATOMIC_NULL)
            else:
                buf.write_uint8(TNS_NULL_LENGTH_INDICATOR)
        elif ora_type_num in (ORA_TYPE_NUM_CHAR, ORA_TYPE_NUM_VARCHAR):
            if metadata.dbtype._csfrm == CS_FORM_IMPLICIT:
                temp_bytes = (<str> value).encode()
            else:
                temp_bytes = (<str> value).encode(ENCODING_UTF16)
            buf.write_bytes_with_length(temp_bytes)
        elif ora_type_num == ORA_TYPE_NUM_NUMBER:
            temp_bytes = (<str> cpython.PyObject_Str(value)).encode()
            buf.write_oracle_number(temp_bytes)
        elif ora_type_num == ORA_TYPE_NUM_BINARY_INTEGER:
            buf.write_uint8(4)
            buf.write_uint32be(<uint32_t> value)
        elif ora_type_num == ORA_TYPE_NUM_RAW:
            buf.write_bytes_with_length(value)
        elif ora_type_num == ORA_TYPE_NUM_BINARY_DOUBLE:
            buf.write_binary_double(value)
        elif ora_type_num == ORA_TYPE_NUM_BINARY_FLOAT:
            buf.write_binary_float(value)
        elif ora_type_num == ORA_TYPE_NUM_BOOLEAN:
            buf.write_uint8(4)
            buf.write_uint32be(value)
        elif ora_type_num in (ORA_TYPE_NUM_DATE, ORA_TYPE_NUM_TIMESTAMP,
                              ORA_TYPE_NUM_TIMESTAMP_TZ,
                              ORA_TYPE_NUM_TIMESTAMP_LTZ):
            buf.write_oracle_date(value, metadata.dbtype._buffer_size_factor)
        elif ora_type_num in (ORA_TYPE_NUM_CLOB, ORA_TYPE_NUM_BLOB):
            lob_impl = <BaseThinLobImpl> value._impl
            buf.write_bytes_with_length(lob_impl._locator)
        elif ora_type_num == ORA_TYPE_NUM_OBJECT:
            obj_impl = value._impl
            if self.type.is_collection or obj_impl.type.is_collection:
                temp_bytes = obj_impl._get_packed_data()
                buf.write_bytes_with_length(temp_bytes)
            else:
                obj_impl._pack_data(buf)
        else:
            errors._raise_err(errors.ERR_DB_TYPE_NOT_SUPPORTED,
                              name=metadata.dbtype.name)

    cdef int _unpack_data(self) except -1:
        """
        Unpacks the packed data into a dictionary of Python values.
        """
        cdef DbObjectPickleBuffer buf
        buf = DbObjectPickleBuffer.__new__(DbObjectPickleBuffer)
        buf._populate_from_bytes(self.packed_data)
        buf.read_header(&self.image_flags, &self.image_version)
        self._unpack_data_from_buf(buf)
        self.packed_data = None

    cdef int _unpack_data_from_buf(self, DbObjectPickleBuffer buf) except -1:
        """
        Unpacks the data from the buffer into Python values.
        """
        cdef:
            dict unpacked_attrs = {}, unpacked_assoc_array = None
            ThinDbObjectTypeImpl typ_impl = self.type
            list unpacked_array = None
            ThinDbObjectAttrImpl attr
            uint32_t num_elements, i
            int32_t assoc_index
            object value
        if typ_impl.is_collection:
            if typ_impl.collection_type == TNS_OBJ_PLSQL_INDEX_TABLE:
                unpacked_assoc_array = {}
            else:
                unpacked_array = []
            buf.skip_raw_bytes(1)           # collection flags
            buf.read_length(&num_elements)
            for i in range(num_elements):
                if typ_impl.collection_type == TNS_OBJ_PLSQL_INDEX_TABLE:
                    buf.read_int32be(&assoc_index)
                value = self._unpack_value(buf, typ_impl.element_metadata)
                if typ_impl.collection_type == TNS_OBJ_PLSQL_INDEX_TABLE:
                    unpacked_assoc_array[assoc_index] = value
                else:
                    unpacked_array.append(value)
        else:
            unpacked_attrs = {}
            for attr in typ_impl.attrs:
                value = self._unpack_value(buf, attr)
                unpacked_attrs[attr.name] = value
        self.unpacked_attrs = unpacked_attrs
        self.unpacked_array = unpacked_array
        self.unpacked_assoc_array = unpacked_assoc_array

    cdef object _unpack_value(self, DbObjectPickleBuffer buf,
                              OracleMetadata metadata):
        """
        Unpacks a single value and returns it.
        """
        cdef:
            uint8_t ora_type_num = metadata.dbtype._ora_type_num
            uint8_t csfrm = metadata.dbtype._csfrm
            DbObjectPickleBuffer xml_buf
            bint is_null, is_collection
            BaseThinConnImpl conn_impl
            ThinDbObjectImpl obj_impl
            BaseThinLobImpl lob_impl
            OracleData data
            bytes locator
            type cls
        if ora_type_num in (ORA_TYPE_NUM_CLOB,
                              ORA_TYPE_NUM_BLOB,
                              ORA_TYPE_NUM_BFILE):
            conn_impl = self.type._conn_impl
            locator = buf.read_bytes()
            if locator is None:
                return None
            lob_impl = conn_impl._create_lob_impl(metadata.dbtype, locator)
            cls = PY_TYPE_ASYNC_LOB \
                    if conn_impl._protocol._transport._is_async \
                    else PY_TYPE_LOB
            return cls._from_impl(lob_impl)
        elif ora_type_num == ORA_TYPE_NUM_OBJECT:
            is_collection = \
                    metadata.objtype.is_collection or self.type.is_collection
            buf.get_is_atomic_null(is_collection, &is_null)
            if is_null:
                return None
            if metadata.objtype is None:
                xml_buf = DbObjectPickleBuffer.__new__(DbObjectPickleBuffer)
                xml_buf._populate_from_bytes(buf.read_bytes())
                return xml_buf.read_xmltype(self.type._conn_impl)
            obj_impl = ThinDbObjectImpl.__new__(ThinDbObjectImpl)
            obj_impl.type = metadata.objtype
            if is_collection:
                obj_impl.packed_data = buf.read_bytes()
            else:
                obj_impl._unpack_data_from_buf(buf)
            return PY_TYPE_DB_OBJECT._from_impl(obj_impl)
        buf.read_oracle_data(metadata, &data, from_dbobject=True)
        if metadata.dbtype._csfrm == CS_FORM_NCHAR:
            conn_impl = self.type._conn_impl
            conn_impl._protocol._caps._check_ncharset_id()
        return convert_oracle_data_to_python(metadata, metadata, &data,
                                             encoding_errors=NULL,
                                             from_dbobject=True)

    def append_checked(self, object value):
        """
        Internal method for appending a value to a collection object.
        """
        cdef:
            ThinDbObjectTypeImpl typ_impl
            int32_t new_index
        self._ensure_unpacked()
        if self.unpacked_array is not None:
            typ_impl = self.type
            if typ_impl.max_num_elements > 0 \
                    and len(self.unpacked_array) >= typ_impl.max_num_elements:
                errors._raise_err(errors.ERR_INVALID_COLL_INDEX_SET,
                                  index=len(self.unpacked_array),
                                  min_index=0,
                                  max_index=typ_impl.max_num_elements - 1)
            self.unpacked_array.append(value)
        else:
            self._ensure_assoc_keys()
            new_index = self.unpacked_assoc_keys[-1] + 1 \
                    if self.unpacked_assoc_keys else 0
            self.unpacked_assoc_array[new_index] = value
            self.unpacked_assoc_keys.append(new_index)

    def copy(self):
        """
        Internal method for creating a copy of an object.
        """
        cdef ThinDbObjectImpl copied_impl
        copied_impl = ThinDbObjectImpl.__new__(ThinDbObjectImpl)
        copied_impl.type = self.type
        copied_impl.flags = self.flags
        copied_impl.image_flags = self.image_flags
        copied_impl.image_version = self.image_version
        copied_impl.toid = self.toid
        copied_impl.packed_data = self.packed_data
        copied_impl.num_elements = self.num_elements
        if self.unpacked_attrs is not None:
            copied_impl.unpacked_attrs = self.unpacked_attrs.copy()
        if self.unpacked_array is not None:
            copied_impl.unpacked_array = list(self.unpacked_array)
        return copied_impl

    def delete_by_index(self, int32_t index):
        """
        Internal method for deleting an entry from a collection that is indexed
        by integers.
        """
        self._ensure_unpacked()
        if self.unpacked_array is not None:
            del self.unpacked_array[index]
        else:
            self.unpacked_assoc_keys = None
            del self.unpacked_assoc_array[index]

    def exists_by_index(self, int32_t index):
        """
        Internal method for determining if an entry exists in a collection that
        is indexed by integers.
        """
        self._ensure_unpacked()
        if self.unpacked_array is not None:
            return index >= 0 and index < len(self.unpacked_array)
        else:
            return index in self.unpacked_assoc_array

    def get_attr_value(self, ThinDbObjectAttrImpl attr):
        """
        Internal method for getting an attribute value.
        """
        self._ensure_unpacked()
        return self.unpacked_attrs[attr.name]

    def get_element_by_index(self, int32_t index):
        """
        Internal method for getting an entry from a collection that is indexed
        by integers.
        """
        self._ensure_unpacked()
        try:
            if self.unpacked_array is not None:
                return self.unpacked_array[index]
            else:
                return self.unpacked_assoc_array[index]
        except (KeyError, IndexError):
            errors._raise_err(errors.ERR_INVALID_COLL_INDEX_GET, index=index)

    def get_first_index(self):
        """
        Internal method for getting the first index from a collection that is
        indexed by integers.
        """
        self._ensure_unpacked()
        if self.unpacked_array:
            return 0
        elif self.unpacked_assoc_array:
            self._ensure_assoc_keys()
            return self.unpacked_assoc_keys[0]

    def get_last_index(self):
        """
        Internal method for getting the last index from a collection that is
        indexed by integers.
        """
        self._ensure_unpacked()
        if self.unpacked_array:
            return len(self.unpacked_array) - 1
        elif self.unpacked_assoc_array:
            self._ensure_assoc_keys()
            return self.unpacked_assoc_keys[-1]

    def get_next_index(self, int32_t index):
        """
        Internal method for getting the next index from a collection that is
        indexed by integers.
        """
        cdef int32_t i
        self._ensure_unpacked()
        if self.unpacked_array:
            if index + 1 < len(self.unpacked_array):
                return index + 1
        elif self.unpacked_assoc_array:
            self._ensure_assoc_keys()
            for i in self.unpacked_assoc_keys:
                if i > index:
                    return i

    def get_prev_index(self, int32_t index):
        """
        Internal method for getting the next index from a collection that is
        indexed by integers.
        """
        self._ensure_unpacked()
        if self.unpacked_array:
            if index > 0:
                return index - 1
        elif self.unpacked_assoc_array:
            self._ensure_assoc_keys()
            for i in reversed(self.unpacked_assoc_keys):
                if i < index:
                    return i

    def get_size(self):
        """
        Internal method for getting the size of a collection.
        """
        self._ensure_unpacked()
        if self.unpacked_array is not None:
            return len(self.unpacked_array)
        else:
            return len(self.unpacked_assoc_array)

    def set_attr_value_checked(self, ThinDbObjectAttrImpl attr, object value):
        """
        Internal method for setting an attribute value.
        """
        self._ensure_unpacked()
        self.unpacked_attrs[attr.name] = value

    def set_element_by_index_checked(self, int32_t index, object value):
        """
        Internal method for setting an entry in a collection that is indexed by
        integers.
        """
        self._ensure_unpacked()
        if self.unpacked_array is not None:
            try:
                self.unpacked_array[index] = value
            except IndexError:
                max_index = max(len(self.unpacked_array) - 1, 0)
                errors._raise_err(errors.ERR_INVALID_COLL_INDEX_SET,
                                  index=index, min_index=0,
                                  max_index=max_index)
        else:
            if index not in self.unpacked_assoc_array:
                self.unpacked_assoc_keys = None
            self.unpacked_assoc_array[index] = value

    def trim(self, int32_t num_to_trim):
        """
        Internal method for trimming a number of entries from a collection.
        """
        self._ensure_unpacked()
        if num_to_trim > 0:
            self.unpacked_array = self.unpacked_array[:-num_to_trim]


cdef class ThinDbObjectAttrImpl(BaseDbObjectAttrImpl):
    cdef:
        bytes oid


cdef class ThinDbObjectTypeImpl(BaseDbObjectTypeImpl):
    cdef:
        uint8_t collection_type, collection_flags, version
        uint32_t max_num_elements
        bint is_row_type
        bint is_xml_type
        bytes oid

    def create_new_object(self):
        """
        Internal method for creating a new object.
        """
        cdef ThinDbObjectImpl obj_impl
        obj_impl = ThinDbObjectImpl.__new__(ThinDbObjectImpl)
        obj_impl.type = self
        obj_impl.toid = b'\x00\x22' + \
                bytes([TNS_OBJ_NON_NULL_OID, TNS_OBJ_HAS_EXTENT_OID]) + \
                self.oid + TNS_EXTENT_OID
        obj_impl.flags = TNS_OBJ_TOP_LEVEL
        obj_impl.image_flags = TNS_OBJ_IS_VERSION_81
        obj_impl.image_version = TNS_OBJ_IMAGE_VERSION
        obj_impl.unpacked_attrs = {}
        if self.is_collection:
            obj_impl.image_flags |= TNS_OBJ_IS_COLLECTION
            if self.collection_type == TNS_OBJ_PLSQL_INDEX_TABLE:
                obj_impl.unpacked_assoc_array = {}
            else:
                obj_impl.unpacked_array = []
        else:
            obj_impl.image_flags |= TNS_OBJ_NO_PREFIX_SEG
            for attr in self.attrs:
                obj_impl.unpacked_attrs[attr.name] = None
        return obj_impl



# -----------------------------------------------------------------------------
# Copyright (c) 2020, 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
1100 - Module for testing connections
"""

import random
import string
import threading
import time
import unittest

import oracledb
import test_env


class TestCase(test_env.BaseTestCase):
    requires_connection = False

    def __connect_and_drop(self):
        """
        Connect to the database, perform a query and drop the connection.
        """
        with test_env.get_connection() as conn:
            cursor = conn.cursor()
            cursor.execute("select count(*) from TestNumbers")
            (count,) = cursor.fetchone()
            self.assertEqual(count, 10)

    def __verify_fetched_data(self, connection):
        expected_data = [f"String {i + 1}" for i in range(10)]
        sql = "select StringCol from TestStrings order by IntCol"
        for i in range(5):
            with connection.cursor() as cursor:
                fetched_data = [s for s, in cursor.execute(sql)]
                self.assertEqual(fetched_data, expected_data)

    def __verify_attributes(self, connection, attr_name, value, sql):
        setattr(connection, attr_name, value)
        cursor = connection.cursor()
        cursor.execute(sql)
        (result,) = cursor.fetchone()
        self.assertEqual(result, value, f"{attr_name} value mismatch")

    def __verify_connect_arg(self, arg_name, arg_value, sql):
        args = {}
        args[arg_name] = arg_value
        conn = test_env.get_connection(**args)
        cursor = conn.cursor()
        cursor.execute(sql)
        (fetched_value,) = cursor.fetchone()
        self.assertEqual(fetched_value, arg_value)

    def test_1100(self):
        "1100 - simple connection to database"
        conn = test_env.get_connection()
        self.assertEqual(
            conn.username, test_env.get_main_user(), "user name differs"
        )
        self.assertEqual(
            conn.dsn, test_env.get_connect_string(), "dsn differs"
        )
        self.assertEqual(conn.thin, test_env.get_is_thin())

    @unittest.skipIf(test_env.get_is_drcp(), "not supported with DRCP")
    def test_1101(self):
        "1101 - test use of application context"
        namespace = "CLIENTCONTEXT"
        app_context_entries = [
            (namespace, "ATTR1", "VALUE1"),
            (namespace, "ATTR2", "VALUE2"),
            (namespace, "ATTR3", "VALUE3"),
        ]
        conn = test_env.get_connection(appcontext=app_context_entries)
        cursor = conn.cursor()
        for namespace, name, value in app_context_entries:
            cursor.execute(
                "select sys_context(:1, :2) from dual", (namespace, name)
            )
            (actual_value,) = cursor.fetchone()
            self.assertEqual(actual_value, value)

    def test_1102(self):
        "1102 - test invalid use of application context"
        self.assertRaises(
            TypeError,
            test_env.get_connection,
            appcontext=[("userenv", "action")],
        )

    def test_1103(self):
        "1103 - test connection end-to-end tracing attributes"
        conn = test_env.get_connection()
        if test_env.get_client_version() >= (
            12,
            1,
        ) and not self.is_on_oracle_cloud(conn):
            sql = """select dbop_name from v$sql_monitor
                     where sid = sys_context('userenv', 'sid')
                     and status = 'EXECUTING'"""
            self.__verify_attributes(conn, "dbop", "oracledb_dbop", sql)
        sql = "select sys_context('userenv', 'action') from dual"
        self.__verify_attributes(conn, "action", "oracledb_Action", sql)
        self.__verify_attributes(conn, "action", None, sql)
        sql = "select sys_context('userenv', 'module') from dual"
        self.__verify_attributes(conn, "module", "oracledb_Module", sql)
        self.__verify_attributes(conn, "module", None, sql)
        sql = "select sys_context('userenv', 'client_info') from dual"
        self.__verify_attributes(conn, "clientinfo", "oracledb_cinfo", sql)
        self.__verify_attributes(conn, "clientinfo", None, sql)
        sql = "select sys_context('userenv', 'client_identifier') from dual"
        self.__verify_attributes(
            conn, "client_identifier", "oracledb_cid", sql
        )
        self.__verify_attributes(conn, "client_identifier", None, sql)
        if not test_env.get_is_thin():
            sql = """select ecid from v$session
                     where sid = sys_context('userenv', 'sid')"""
            self.__verify_attributes(conn, "econtext_id", "oracledb_ecid", sql)
            self.__verify_attributes(conn, "econtext_id", None, sql)

    def test_1104(self):
        "1104 - test use of autocommit"
        conn = test_env.get_connection()
        cursor = conn.cursor()
        other_conn = test_env.get_connection()
        other_cursor = other_conn.cursor()
        cursor.execute("truncate table TestTempTable")
        cursor.execute("insert into TestTempTable (IntCol) values (1)")
        other_cursor.execute("select IntCol from TestTempTable")
        self.assertEqual(other_cursor.fetchall(), [])
        conn.autocommit = True
        cursor.execute("insert into TestTempTable (IntCol) values (2)")
        other_cursor.execute(
            "select IntCol from TestTempTable order by IntCol"
        )
        self.assertEqual(other_cursor.fetchall(), [(1,), (2,)])

    def test_1105(self):
        "1105 - connection to database with bad connect string"
        with self.assertRaisesFullCode(
            "DPY-4000", "DPY-4026", "DPY-4027", "ORA-12154"
        ):
            oracledb.connect("not a valid connect string!!")
        with self.assertRaisesFullCode("DPY-4000", "DPY-4001"):
            dsn = (
                test_env.get_main_user() + "@" + test_env.get_connect_string()
            )
            oracledb.connect(dsn)

    def test_1106(self):
        "1106 - connection to database with bad password"
        with self.assertRaisesFullCode("ORA-01017"):
            test_env.get_connection(
                password=test_env.get_main_password() + "X"
            )

    @unittest.skipIf(test_env.get_is_drcp(), "not supported with DRCP")
    def test_1107(self):
        "1107 - test changing password"
        conn = test_env.get_connection()
        if self.is_on_oracle_cloud(conn):
            self.skipTest("passwords on Oracle Cloud are strictly controlled")
        sys_random = random.SystemRandom()
        new_password = "".join(
            sys_random.choice(string.ascii_letters) for i in range(20)
        )
        conn.changepassword(test_env.get_main_password(), new_password)
        conn = test_env.get_connection(password=new_password)
        conn.changepassword(new_password, test_env.get_main_password())

    @unittest.skipIf(test_env.get_is_drcp(), "not supported with DRCP")
    def test_1108(self):
        "1108 - test changing password to an invalid value"
        conn = test_env.get_connection()
        if self.is_on_oracle_cloud(conn):
            self.skipTest("passwords on Oracle Cloud are strictly controlled")
        new_password = "1" * 1500
        with self.assertRaisesFullCode("ORA-01017", "ORA-00988"):
            conn.changepassword(test_env.get_main_password(), new_password)
        with self.assertRaisesFullCode("ORA-01017", "ORA-00988", "ORA-28008"):
            conn.changepassword("incorrect old password", new_password)

    @unittest.skipIf(test_env.get_is_drcp(), "not supported with DRCP")
    def test_1109(self):
        "1109 - test connecting with password containing / and @ symbols"
        conn = test_env.get_connection()
        if self.is_on_oracle_cloud(conn):
            self.skipTest("passwords on Oracle Cloud are strictly controlled")
        sys_random = random.SystemRandom()
        chars = list(
            sys_random.choice(string.ascii_letters) for i in range(20)
        )
        chars[4] = "/"
        chars[8] = "@"
        new_password = "".join(chars)
        conn.changepassword(test_env.get_main_password(), new_password)
        try:
            test_env.get_connection(password=new_password)
        finally:
            conn.changepassword(new_password, test_env.get_main_password())

    def test_1110(self):
        "1110 - confirm an exception is raised after closing a connection"
        conn = test_env.get_connection()
        conn.close()
        with self.assertRaisesFullCode("DPY-1001"):
            conn.rollback()

    @unittest.skipIf(test_env.get_is_thin(), "not relevant for thin mode")
    def test_1111(self):
        "1111 - test creating a connection using a handle"
        conn = test_env.get_connection()
        cursor = conn.cursor()
        cursor.execute("truncate table TestTempTable")
        int_value = random.randint(1, 32768)
        cursor.execute(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (:val, null)
            """,
            val=int_value,
        )
        conn2 = oracledb.connect(handle=conn.handle)
        cursor = conn2.cursor()
        cursor.execute("select IntCol from TestTempTable")
        (fetched_int_value,) = cursor.fetchone()
        self.assertEqual(fetched_int_value, int_value)

        cursor.close()
        with self.assertRaisesFullCode("DPI-1034"):
            conn2.close()
        conn.close()

    def test_1112(self):
        "1112 - connection version is a string"
        conn = test_env.get_connection()
        self.assertIsInstance(conn.version, str)

    def test_1113(self):
        "1113 - connection rolls back before close"
        conn = test_env.get_connection()
        cursor = conn.cursor()
        cursor.execute("truncate table TestTempTable")
        other_conn = test_env.get_connection()
        other_cursor = other_conn.cursor()
        other_cursor.execute("insert into TestTempTable (IntCol) values (1)")
        other_cursor.close()
        other_conn.close()
        cursor.execute("select count(*) from TestTempTable")
        (count,) = cursor.fetchone()
        self.assertEqual(count, 0)

    def test_1114(self):
        "1114 - connection rolls back before destruction"
        conn = test_env.get_connection()
        cursor = conn.cursor()
        cursor.execute("truncate table TestTempTable")
        other_conn = test_env.get_connection()
        other_cursor = other_conn.cursor()
        other_cursor.execute("insert into TestTempTable (IntCol) values (1)")
        del other_cursor
        del other_conn
        cursor.execute("select count(*) from TestTempTable")
        (count,) = cursor.fetchone()
        self.assertEqual(count, 0)

    def test_1115(self):
        "1115 - multiple connections to database with multiple threads"
        threads = []
        for i in range(20):
            thread = threading.Thread(None, self.__connect_and_drop)
            threads.append(thread)
            thread.start()
            time.sleep(0.1)
        for thread in threads:
            thread.join()

    def test_1116(self):
        "1116 - test string format of connection"
        conn = test_env.get_connection()
        expected_value = "<oracledb.Connection to %s@%s>" % (
            test_env.get_main_user(),
            test_env.get_connect_string(),
        )
        self.assertEqual(str(conn), expected_value)

    def test_1117(self):
        "1117 - test context manager - close"
        with test_env.get_connection() as conn:
            cursor = conn.cursor()
            cursor.execute("truncate table TestTempTable")
            cursor.execute("insert into TestTempTable (IntCol) values (1)")
            conn.commit()
            cursor.execute("insert into TestTempTable (IntCol) values (2)")
        with self.assertRaisesFullCode("DPY-1001"):
            conn.ping()
        conn = test_env.get_connection()
        cursor = conn.cursor()
        cursor.execute("select count(*) from TestTempTable")
        (count,) = cursor.fetchone()
        self.assertEqual(count, 1)

    def test_1118(self):
        "1118 - test connection attribute values"
        conn = test_env.get_connection()
        if test_env.get_client_version() >= (12, 1):
            self.assertEqual(conn.ltxid, b"")
        self.assertFalse(conn.autocommit)
        conn.autocommit = True
        self.assertTrue(conn.autocommit)
        self.assertIsNone(conn.current_schema)
        conn.current_schema = "test_schema"
        self.assertEqual(conn.current_schema, "test_schema")
        self.assertIsNone(conn.edition)
        conn.external_name = "test_external"
        self.assertEqual(conn.external_name, "test_external")
        conn.internal_name = "test_internal"
        self.assertEqual(conn.internal_name, "test_internal")
        if conn.max_identifier_length is not None:
            self.assertIsInstance(conn.max_identifier_length, int)
        conn.stmtcachesize = 30
        self.assertEqual(conn.stmtcachesize, 30)
        self.assertRaises(TypeError, conn.stmtcachesize, 20.5)
        self.assertRaises(TypeError, conn.stmtcachesize, "value")
        self.assertIsNone(conn.warning)

    def test_1119(self):
        "1119 - test closed connection attribute values"
        conn = test_env.get_connection()
        conn.close()
        attr_names = [
            "current_schema",
            "edition",
            "external_name",
            "internal_name",
            "stmtcachesize",
            "warning",
        ]
        if test_env.get_client_version() >= (12, 1):
            attr_names.append("ltxid")
        for name in attr_names:
            with self.assertRaisesFullCode("DPY-1001"):
                getattr(conn, name)

    def test_1120(self):
        "1120 - test connection ping makes a round trip"
        self.conn = test_env.get_connection()
        self.setup_round_trip_checker()
        self.conn.ping()
        self.assertRoundTrips(1)

    @unittest.skipIf(
        test_env.get_is_thin(),
        "thin mode doesn't support two-phase commit yet",
    )
    def test_1121(self):
        "1121 - test begin, prepare, cancel transaction"
        conn = test_env.get_connection()
        cursor = conn.cursor()
        cursor.execute("truncate table TestTempTable")
        conn.begin(10, "trxnId", "branchId")
        self.assertFalse(conn.prepare())
        conn.begin(10, "trxnId", "branchId")
        cursor.execute(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (1, 'tesName')
            """
        )
        self.assertTrue(conn.prepare())
        conn.cancel()
        conn.rollback()
        cursor.execute("select count(*) from TestTempTable")
        (count,) = cursor.fetchone()
        self.assertEqual(count, 0)

    @unittest.skipIf(
        test_env.get_is_thin(),
        "thin mode doesn't support two-phase commit yet",
    )
    def test_1122(self):
        "1122 - test multiple transactions on the same connection"
        conn = test_env.get_connection()
        with conn.cursor() as cursor:
            cursor.execute("truncate table TestTempTable")

        id_ = random.randint(0, 2**128)
        xid = (0x1234, "%032x" % id_, "%032x" % 9)
        conn.begin(*xid)
        with conn.cursor() as cursor:
            cursor.execute(
                """
                insert into TestTempTable (IntCol, StringCol1)
                values (1, 'tesName')
                """
            )
            self.assertTrue(conn.prepare())
            conn.commit()

        for begin_trans in (True, False):
            val = 3
            if begin_trans:
                conn.begin()
                val = 2
            with conn.cursor() as cursor:
                cursor.execute(
                    """
                    insert into TestTempTable (IntCol, StringCol1)
                    values (:int_val, 'tesName')
                    """,
                    int_val=val,
                )
                conn.commit()

        expected_rows = [(1, "tesName"), (2, "tesName"), (3, "tesName")]
        with conn.cursor() as cursor:
            cursor.execute("select IntCol, StringCol1 from TestTempTable")
            self.assertEqual(cursor.fetchall(), expected_rows)

    @unittest.skipIf(
        test_env.get_is_thin(),
        "thin mode doesn't support two-phase commit yet",
    )
    def test_1123(self):
        "1123 - test multiple global transactions on the same connection"
        conn = test_env.get_connection()
        with conn.cursor() as cursor:
            cursor.execute("truncate table TestTempTable")

        id_ = random.randint(0, 2**128)
        xid = (0x1234, "%032x" % id_, "%032x" % 9)
        conn.begin(*xid)
        with conn.cursor() as cursor:
            cursor.execute(
                """
                insert into TestTempTable (IntCol, StringCol1)
                values (1, 'tesName')
                """
            )
            self.assertTrue(conn.prepare())
            conn.commit()

        for begin_trans in (True, False):
            val = 3
            if begin_trans:
                conn.begin()
                val = 2
            with conn.cursor() as cursor:
                cursor.execute(
                    """
                    insert into TestTempTable (IntCol, StringCol1)
                    values (:int_val, 'tesName')
                    """,
                    int_val=val,
                )
                conn.commit()

        id2_ = random.randint(0, 2**128)
        xid2 = (0x1234, "%032x" % id2_, "%032x" % 9)
        conn.begin(*xid2)
        with conn.cursor() as cursor:
            cursor.execute(
                """
                insert into TestTempTable (IntCol, StringCol1)
                values (4, 'tesName')
                """
            )
            self.assertTrue(conn.prepare())
            conn.commit()

        expected_rows = [
            (1, "tesName"),
            (2, "tesName"),
            (3, "tesName"),
            (4, "tesName"),
        ]
        with conn.cursor() as cursor:
            cursor.execute("select IntCol, StringCol1 from TestTempTable")
            self.assertEqual(cursor.fetchall(), expected_rows)

    @unittest.skipIf(
        test_env.get_is_thin(),
        "thin mode doesn't support two-phase commit yet",
    )
    def test_1124(self):
        "1124 - test creating global txn after a local txn"
        conn = test_env.get_connection()
        with conn.cursor() as cursor:
            cursor.execute("truncate table TestTempTable")

        with conn.cursor() as cursor:
            cursor.execute(
                """
                insert into TestTempTable (IntCol, StringCol1)
                values (2, 'tesName')
                """
            )

        id_ = random.randint(0, 2**128)
        xid = (0x1234, "%032x" % id_, "%032x" % 9)
        with self.assertRaisesFullCode("ORA-24776"):
            conn.begin(*xid)

    def test_1125(self):
        "1125 - single connection to database with multiple threads"
        with test_env.get_connection() as conn:
            threads = [
                threading.Thread(
                    target=self.__verify_fetched_data, args=(conn,)
                )
                for i in range(3)
            ]
            for t in threads:
                t.start()
            for t in threads:
                t.join()

    def test_1126(self):
        "1126 - test connection cancel"
        conn = test_env.get_connection()
        sleep_proc_name = test_env.get_sleep_proc_name()

        def perform_cancel():
            time.sleep(0.1)
            conn.cancel()

        thread = threading.Thread(target=perform_cancel)
        thread.start()
        try:
            with conn.cursor() as cursor:
                self.assertRaises(
                    oracledb.OperationalError,
                    cursor.callproc,
                    sleep_proc_name,
                    [2],
                )
        finally:
            thread.join()
        with conn.cursor() as cursor:
            cursor.execute("select user from dual")
            (user,) = cursor.fetchone()
            self.assertEqual(user, test_env.get_main_user().upper())

    @unittest.skipIf(test_env.get_is_drcp(), "not supported with DRCP")
    def test_1127(self):
        "1127 - test changing password during connect"
        conn = test_env.get_connection()
        if self.is_on_oracle_cloud(conn):
            self.skipTest("passwords on Oracle Cloud are strictly controlled")
        sys_random = random.SystemRandom()
        new_password = "".join(
            sys_random.choice(string.ascii_letters) for i in range(20)
        )
        conn = test_env.get_connection(newpassword=new_password)
        conn = test_env.get_connection(password=new_password)
        conn.changepassword(new_password, test_env.get_main_password())

    def test_1128(self):
        "1128 - test use of autocommit during reexecute"
        sql = "insert into TestTempTable (IntCol, StringCol1) values (:1, :2)"
        data_to_insert = [(1, "Test String #1"), (2, "Test String #2")]
        conn = test_env.get_connection()
        cursor = conn.cursor()
        other_conn = test_env.get_connection()
        other_cursor = other_conn.cursor()
        cursor.execute("truncate table TestTempTable")
        cursor.execute(sql, data_to_insert[0])
        other_cursor.execute("select IntCol, StringCol1 from TestTempTable")
        self.assertEqual(other_cursor.fetchall(), [])
        conn.autocommit = True
        cursor.execute(sql, data_to_insert[1])
        other_cursor.execute("select IntCol, StringCol1 from TestTempTable")
        self.assertEqual(other_cursor.fetchall(), data_to_insert)

    def test_1129(self):
        "1129 - test current_schema is set properly"
        conn = test_env.get_connection()
        self.assertIsNone(conn.current_schema)

        user = test_env.get_main_user().upper()
        proxy_user = test_env.get_proxy_user().upper()
        cursor = conn.cursor()
        cursor.execute(f"alter session set current_schema={proxy_user}")
        self.assertEqual(conn.current_schema, proxy_user)

        conn.current_schema = user
        self.assertEqual(conn.current_schema, user)

        cursor.execute(
            "select sys_context('userenv', 'current_schema') from dual"
        )
        (result,) = cursor.fetchone()
        self.assertEqual(result, user)

    def test_1130(self):
        "1130 - test dbms_output package"
        conn = test_env.get_connection()
        cursor = conn.cursor()
        test_string = "Testing DBMS_OUTPUT package"
        cursor.callproc("dbms_output.enable")
        cursor.callproc("dbms_output.put_line", [test_string])
        string_var = cursor.var(str)
        number_var = cursor.var(int)
        cursor.callproc("dbms_output.get_line", (string_var, number_var))
        self.assertEqual(string_var.getvalue(), test_string)

    @unittest.skipIf(
        not test_env.get_is_thin() and test_env.get_client_version() < (18, 1),
        "unsupported client",
    )
    def test_1131(self):
        "1131 - test connection call_timeout"
        conn = test_env.get_connection()
        conn.call_timeout = 500  # milliseconds
        self.assertEqual(conn.call_timeout, 500)
        with self.assertRaisesFullCode("DPY-4011", "DPY-4024"):
            conn.cursor().callproc(test_env.get_sleep_proc_name(), [2])

    def test_1132(self):
        "1132 - test Connection repr()"

        class MyConnection(oracledb.Connection):
            pass

        conn = test_env.get_connection(conn_class=MyConnection)
        qual_name = conn.__class__.__qualname__
        expected_value = (
            f"<{__name__}.{qual_name} to {conn.username}@{conn.dsn}>"
        )
        self.assertEqual(repr(conn), expected_value)

        conn.close()
        expected_value = f"<{__name__}.{qual_name} disconnected>"
        self.assertEqual(repr(conn), expected_value)

    def test_1133(self):
        "1133 - test getting write-only attributes"
        conn = test_env.get_connection()
        with self.assertRaises(AttributeError):
            conn.action
        with self.assertRaises(AttributeError):
            conn.dbop
        with self.assertRaises(AttributeError):
            conn.clientinfo
        with self.assertRaises(AttributeError):
            conn.econtext_id
        with self.assertRaises(AttributeError):
            conn.module
        with self.assertRaises(AttributeError):
            conn.client_identifier

    def test_1134(self):
        "1134 - test error for invalid type for params and pool"
        pool = test_env.get_pool()
        pool.close()
        with self.assertRaisesFullCode("DPY-1002"):
            test_env.get_connection(pool=pool)
        self.assertRaises(
            TypeError,
            test_env.get_connection,
            pool="This isn't an instance of a pool",
        )
        with self.assertRaisesFullCode("DPY-2025"):
            oracledb.connect(params={"number": 7})

    def test_1135(self):
        "1135 - test connection instance name"
        conn = test_env.get_connection()
        cursor = conn.cursor()
        cursor.execute(
            """
            select upper(sys_context('userenv', 'instance_name'))
            from dual
            """
        )
        (instance_name,) = cursor.fetchone()
        self.assertEqual(conn.instance_name.upper(), instance_name)

    @unittest.skipIf(
        test_env.get_client_version() < (18, 1), "not supported on this client"
    )
    def test_1136(self):
        "1136 - test deprecated attributes"
        conn = test_env.get_connection()
        conn.callTimeout = 500
        self.assertEqual(conn.callTimeout, 500)

    @unittest.skipIf(test_env.get_is_drcp(), "not supported with DRCP")
    @unittest.skipIf(
        test_env.get_server_version() < (23, 0)
        or test_env.get_client_version() < (23, 0),
        "unsupported client/server",
    )
    def test_1137(self):
        "1137 - test maximum allowed length for password"
        conn = test_env.get_connection()
        if self.is_on_oracle_cloud(conn):
            self.skipTest("passwords on Oracle Cloud are strictly controlled")

        original_password = test_env.get_main_password()
        new_password_32 = "a" * 32
        conn.changepassword(original_password, new_password_32)
        conn = test_env.get_connection(password=new_password_32)

        new_password_1024 = "a" * 1024
        conn.changepassword(new_password_32, new_password_1024)
        conn = test_env.get_connection(password=new_password_1024)
        conn.changepassword(new_password_1024, original_password)

        new_password_1025 = "a" * 1025
        with self.assertRaisesFullCode("ORA-28218", "ORA-00972"):
            conn.changepassword(original_password, new_password_1025)

    def test_1138(self):
        "1138 - test getting db_name"
        conn = test_env.get_connection()
        cursor = conn.cursor()
        cursor.execute("select name from V$DATABASE")
        (db_name,) = cursor.fetchone()
        self.assertEqual(conn.db_name.upper(), db_name.upper())

    def test_1139(self):
        "1139 - test getting max_open_cursors"
        conn = test_env.get_connection()
        cursor = conn.cursor()
        cursor.execute(
            "select value from V$PARAMETER where name='open_cursors'"
        )
        (max_open_cursors,) = cursor.fetchone()
        self.assertEqual(conn.max_open_cursors, int(max_open_cursors))

    def test_1140(self):
        "1140 - test getting service_name"
        conn = test_env.get_connection()
        cursor = conn.cursor()
        cursor.execute(
            "select sys_context('userenv', 'service_name') from dual"
        )
        (service_name,) = cursor.fetchone()
        self.assertEqual(conn.service_name.upper(), service_name.upper())

    def test_1141(self):
        "1141 - test transaction_in_progress"
        conn = test_env.get_connection()
        self.assertFalse(conn.transaction_in_progress)

        cursor = conn.cursor()
        cursor.execute("truncate table TestTempTable")
        self.assertFalse(conn.transaction_in_progress)

        cursor.execute("insert into TestTempTable (IntCol) values (1)")
        self.assertTrue(conn.transaction_in_progress)

        conn.commit()
        self.assertFalse(conn.transaction_in_progress)

    def test_1142(self):
        "1142 - test getting db_domain"
        conn = test_env.get_connection()
        cursor = conn.cursor()
        cursor.execute("select value from V$PARAMETER where name='db_domain'")
        (db_domain,) = cursor.fetchone()
        self.assertEqual(conn.db_domain, db_domain)

    def test_1143(self):
        "1143 - test connecting with a proxy user"
        proxy_user = test_env.get_proxy_user()
        conn = test_env.get_connection(proxy_user=proxy_user)
        self.assertEqual(conn.username, test_env.get_main_user())
        self.assertEqual(conn.proxy_user, proxy_user)

    @unittest.skipIf(
        not test_env.get_is_thin(), "thick mode doesn't support SDU yet"
    )
    def test_1144(self):
        "1144 - test connection.sdu"
        conn = test_env.get_connection()
        sdu = random.randint(512, conn.sdu)
        conn = test_env.get_connection(sdu=sdu)
        self.assertEqual(conn.sdu, sdu)

    def test_1145(self):
        "1145 - test connection with invalid conn_class"
        with self.assertRaisesFullCode("DPY-2023"):
            test_env.get_connection(conn_class=oracledb.ConnectionPool)

    @unittest.skipUnless(
        test_env.get_is_thin(),
        "thick mode doesn't support program yet",
    )
    def test_1146(self):
        "1146 - test passing program when creating a connection"
        sql = (
            "select program from v$session "
            "where sid = sys_context('userenv', 'sid')"
        )
        self.__verify_connect_arg("program", "newprogram", sql)

    @unittest.skipUnless(
        test_env.get_is_thin(),
        "thick mode doesn't support machine yet",
    )
    def test_1147(self):
        "1147 - test passing machine when creating a connection"
        sql = (
            "select machine from v$session "
            "where sid = sys_context('userenv', 'sid')"
        )
        self.__verify_connect_arg("machine", "newmachine", sql)

    @unittest.skipUnless(
        test_env.get_is_thin(),
        "thick mode doesn't support terminal yet",
    )
    def test_1148(self):
        "1148 - test passing terminal when creating a connection"
        sql = (
            "select terminal from v$session "
            "where sid = sys_context('userenv', 'sid')"
        )
        self.__verify_connect_arg("terminal", "newterminal", sql)

    @unittest.skipUnless(
        test_env.get_is_thin(),
        "thick mode doesn't support osuser yet",
    )
    def test_1149(self):
        "1149 - test passing osuser when creating a connection"
        sql = (
            "select osuser from v$session "
            "where sid = sys_context('userenv', 'sid')"
        )
        self.__verify_connect_arg("osuser", "newosuser", sql)

    def test_1150(self):
        "1150 - test passing driver_name when creating a connection"
        sql = (
            "select distinct client_driver from v$session_connect_info "
            "where sid = sys_context('userenv', 'sid')"
        )
        self.__verify_connect_arg("driver_name", "newdriver", sql)

    @unittest.skipUnless(
        test_env.get_is_thin(), "thick mode doesn't support session_id yet"
    )
    def test_1151(self):
        "1151 - test getting session id"
        conn = test_env.get_connection()
        cursor = conn.cursor()
        cursor.execute("select dbms_debug_jdwp.current_session_id from dual")
        (fetched_value,) = cursor.fetchone()
        self.assertEqual(conn.session_id, fetched_value)

    @unittest.skipUnless(
        test_env.get_is_thin(), "thick mode doesn't support serial_num yet"
    )
    def test_1152(self):
        "1152 - test getting session serial number"
        conn = test_env.get_connection()
        cursor = conn.cursor()
        cursor.execute(
            "select dbms_debug_jdwp.current_session_serial from dual"
        )
        (fetched_value,) = cursor.fetchone()
        self.assertEqual(conn.serial_num, fetched_value)

    @unittest.skipUnless(
        test_env.get_is_thin(),
        "thick mode doesn't support registered protocols",
    )
    def test_1153(self):
        "1153 - test passed params in hook with standalone connection"
        sdu = 4096
        params = test_env.get_connect_params()
        protocol = "proto-test"
        orig_connect_string = test_env.get_connect_string()
        connect_string = f"{protocol}://{orig_connect_string}"

        def hook(passed_protocol, passed_protocol_arg, passed_params):
            self.assertEqual(passed_protocol, protocol)
            self.assertEqual(passed_protocol_arg, orig_connect_string)
            passed_params.parse_connect_string(passed_protocol_arg)
            passed_params.set(sdu=sdu)

        try:
            oracledb.register_protocol(protocol, hook)
            oracledb.connect(dsn=connect_string, params=params)
            self.assertEqual(params.sdu, sdu)
        finally:
            oracledb.register_protocol(protocol, None)

    def test_1154(self):
        "1154 - test altering connection edition"
        conn = test_env.get_connection()
        self.assertIsNone(conn.edition)
        cursor = conn.cursor()
        sql = "select sys_context('USERENV', 'CURRENT_EDITION_NAME') from dual"
        default_edition = "ORA$BASE"
        test_edition = test_env.get_edition_name()
        for edition in [test_edition, default_edition]:
            with self.subTest(edition=edition):
                cursor.execute(f"alter session set edition = {edition}")
                cursor.execute(sql)
                self.assertEqual(conn.edition, cursor.fetchone()[0])
                self.assertEqual(conn.edition, edition.upper())

    def test_1155(self):
        "1155 - test connect() with edition"
        edition = test_env.get_edition_name()
        conn = test_env.get_connection(edition=edition)
        cursor = conn.cursor()
        cursor.execute(
            "select sys_context('USERENV', 'CURRENT_EDITION_NAME') from dual"
        )
        self.assertEqual(cursor.fetchone()[0], edition.upper())
        self.assertEqual(conn.edition, edition)

    def test_1156(self):
        "1156 - test connect() with parameters hook"
        conn = test_env.get_connection()
        orig_stmtcachesize = conn.stmtcachesize
        stmtcachesize = orig_stmtcachesize + 10

        def hook(params):
            params.set(stmtcachesize=stmtcachesize)

        try:
            oracledb.register_params_hook(hook)
            conn = test_env.get_connection()
            self.assertEqual(conn.stmtcachesize, stmtcachesize)
        finally:
            oracledb.unregister_params_hook(hook)

        conn = test_env.get_connection()
        self.assertEqual(conn.stmtcachesize, orig_stmtcachesize)

    def test_1157(self):
        "1157 - test connect() with multiple parameters hooks"

        def hook1(params):
            order.append("first")

        def hook2(params):
            order.append("second")

        def hook3(params):
            order.append("third")

        oracledb.register_params_hook(hook1)
        oracledb.register_params_hook(hook2)
        oracledb.register_params_hook(hook3)
        try:
            order = []
            test_env.get_connection()
            self.assertEqual(order, ["first", "second", "third"])
        finally:
            oracledb.unregister_params_hook(hook1)
            oracledb.unregister_params_hook(hook2)
            oracledb.unregister_params_hook(hook3)

    def test_1158(self):
        "1158 - test error in the middle of a database response"
        conn = test_env.get_connection()
        cursor = conn.cursor()
        cursor.execute("truncate table TestTempTable")
        data = [(i + 1, 2 if i < 1499 else 0) for i in range(1500)]
        cursor.executemany(
            "insert into TestTempTable (IntCol, NumberCol) values (:1, :2)",
            data,
        )
        conn.commit()
        cursor.arraysize = 1500
        with self.assertRaisesFullCode("ORA-01476"):
            cursor.execute(
                """
                select IntCol, 1 / NumberCol
                from TestTempTable
                where IntCol < 1500
                union all
                select IntCol, 1 / NumberCol
                from TestTempTable
                where IntCol = 1500
                """
            )
            cursor.fetchall()


if __name__ == "__main__":
    test_env.run_test_cases()



# -----------------------------------------------------------------------------
# Copyright (c) 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
E2400 - Module for testing Transaction Guard (TG) using asyncio. No special
setup is required but the test suite makes use of debugging packages that are
not intended for normal use. It also creates and drops a service.
"""

import unittest

import oracledb
import test_env


@unittest.skipUnless(
    test_env.get_is_thin(), "asyncio not supported in thick mode"
)
class TestCase(test_env.BaseAsyncTestCase):
    service_name = "oracledb-test-tg-async"
    requires_connection = False
    setup_completed = False

    async def __perform_setup(self):
        """
        Perform setup, if needed.
        """
        if self.__class__.setup_completed:
            return
        user = test_env.get_main_user()
        async with test_env.get_admin_connection(use_async=True) as conn:
            cursor = conn.cursor()
            await cursor.execute(f"grant execute on dbms_tg_dbg to {user}")
            await cursor.execute(f"grant execute on dbms_app_cont to {user}")
            await cursor.execute(
                """
                select count(*) from dba_services
                where name = :name
                """,
                name=self.service_name,
            )
            (count,) = await cursor.fetchone()
            if count > 0:
                try:
                    await cursor.callproc(
                        "dbms_service.start_service", [self.service_name]
                    )
                except Exception as e:
                    if not str(e).startswith("ORA-44305:"):
                        raise
                return
            await cursor.execute(
                f"""
                declare
                    params          dbms_service.svc_parameter_array;
                begin
                    params('COMMIT_OUTCOME') := 'true';
                    params('RETENTION_TIMEOUT') := 604800;
                    dbms_service.create_service('{self.service_name}',
                                                '{self.service_name}', params);
                    dbms_service.start_service('{self.service_name}');
                end;
                """
            )

    async def test_ext_2400(self):
        "E2400 - test standalone connection"
        await self.__perform_setup()
        params = test_env.get_connect_params().copy()
        params.parse_connect_string(test_env.get_connect_string())
        params.set(service_name=self.service_name)
        for arg_name in ("pre_commit", "post_commit"):
            with self.subTest(arg_name=arg_name):
                conn = await oracledb.connect_async(params=params)
                cursor = conn.cursor()
                await cursor.execute("truncate table TestTempTable")
                await cursor.execute(
                    """
                    insert into TestTempTable (IntCol, StringCol1)
                    values (:1, :2)
                    """,
                    [2400, "String for test 2400"],
                )
                full_arg_name = f"dbms_tg_dbg.tg_failpoint_{arg_name}"
                await cursor.execute(
                    f"""
                    begin
                        dbms_tg_dbg.set_failpoint({full_arg_name});
                    end;
                    """
                )
                ltxid = conn.ltxid
                with self.assertRaisesFullCode("DPY-4011"):
                    await conn.commit()
                conn = await oracledb.connect_async(params=params)
                cursor = conn.cursor()
                committed_var = cursor.var(bool)
                completed_var = cursor.var(bool)
                await cursor.callproc(
                    "dbms_app_cont.get_ltxid_outcome",
                    [ltxid, committed_var, completed_var],
                )
                expected_value = arg_name == "post_commit"
                self.assertEqual(committed_var.getvalue(), expected_value)
                self.assertEqual(completed_var.getvalue(), expected_value)

    async def test_ext_2401(self):
        "E2401 - test pooled connection"
        await self.__perform_setup()
        params = test_env.get_pool_params().copy()
        params.parse_connect_string(test_env.get_connect_string())
        params.set(service_name=self.service_name, max=10)
        pool = oracledb.create_pool_async(params=params)
        for arg_name in ("pre_commit", "post_commit"):
            with self.subTest(arg_name=arg_name):
                async with pool.acquire() as conn:
                    cursor = conn.cursor()
                    await cursor.execute("truncate table TestTempTable")
                    await cursor.execute(
                        """
                        insert into TestTempTable (IntCol, StringCol1)
                        values (:1, :2)
                        """,
                        [2400, "String for test 2400"],
                    )
                    full_arg_name = f"dbms_tg_dbg.tg_failpoint_{arg_name}"
                    await cursor.execute(
                        f"""
                        begin
                            dbms_tg_dbg.set_failpoint({full_arg_name});
                        end;
                        """
                    )
                    ltxid = conn.ltxid
                    with self.assertRaisesFullCode("DPY-4011"):
                        await conn.commit()
                async with pool.acquire() as conn:
                    cursor = conn.cursor()
                    committed_var = cursor.var(bool)
                    completed_var = cursor.var(bool)
                    await cursor.callproc(
                        "dbms_app_cont.get_ltxid_outcome",
                        [ltxid, committed_var, completed_var],
                    )
                    expected_value = arg_name == "post_commit"
                    self.assertEqual(committed_var.getvalue(), expected_value)
                    self.assertEqual(completed_var.getvalue(), expected_value)
        await pool.close()


if __name__ == "__main__":
    test_env.run_test_cases()



# -----------------------------------------------------------------------------
# Copyright (c) 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
E1100 - Module for testing external authentication. This requires access to the
orapki executable found in a full database installation and will create and
drop users in the database. The tests here will only run if the has_orapki
value is enabled and, due to limitations in how the Oracle Client libraries
work, can only be run by itself.
"""

import os
import subprocess
import tempfile
import unittest

import oracledb
import test_env


@unittest.skipUnless(
    test_env.get_extended_config_bool("has_orapki"),
    "extended configuration has_orapki is disabled",
)
@unittest.skipIf(
    test_env.get_is_thin(),
    "thin mode doesn't support external authentication yet",
)
class TestCase(test_env.BaseTestCase):
    alias_name = "ext_test_1100"
    user = "ext_test_1100_user"
    requires_connection = False

    @classmethod
    def _build_sqlnet_config(cls):
        """
        Builds the sqlnet.ora configuration file.
        """
        connect_string = test_env.get_connect_string()
        cls.password = test_env.get_random_string()
        subprocess.run(
            [
                "orapki",
                "wallet",
                "create",
                "-wallet",
                cls.temp_dir.name,
                "-auto_login_only",
            ],
            stdout=subprocess.DEVNULL,
        )
        subprocess.run(
            [
                "mkstore",
                "-wrl",
                cls.temp_dir.name,
                "-createCredential",
                cls.alias_name,
                cls.user,
                cls.password,
            ],
            stdout=subprocess.DEVNULL,
        )
        subprocess.run(
            [
                "mkstore",
                "-wrl",
                cls.temp_dir.name,
                "-createCredential",
                connect_string,
                cls.user,
                cls.password,
            ],
            stdout=subprocess.DEVNULL,
        )
        contents = (
            "WALLET_LOCATION=(SOURCE=(METHOD=FILE)(METHOD_DATA="
            + f"(DIRECTORY={cls.temp_dir.name})))\n"
            + "SQLNET.WALLET_OVERRIDE=TRUE"
        )
        file_name = os.path.join(cls.temp_dir.name, "sqlnet.ora")
        with open(file_name, "w") as f:
            f.write(contents)

    @classmethod
    def _build_tnsnames_config(cls):
        """
        Builds the tnsnames.ora configuration file.
        """
        params = oracledb.ConnectParams()
        params.parse_connect_string(test_env.get_connect_string())
        connect_string = params.get_connect_string()
        file_name = os.path.join(cls.temp_dir.name, "tnsnames.ora")
        with open(file_name, "w") as f:
            f.write(f"{cls.alias_name} = {connect_string}")

    @classmethod
    def _build_user(cls):
        """
        Builds the user credentials stored in the supplied wallet.
        """
        admin_conn = test_env.get_admin_connection()
        with admin_conn.cursor() as cursor:
            try:
                cursor.execute(f"drop user {cls.user} cascade")
            except oracledb.DatabaseError:
                pass
            cursor.execute(
                f"create user {cls.user} identified by " + f"{cls.password}"
            )
            cursor.execute(f"grant create session to {cls.user}")

    @classmethod
    def setUpClass(cls):
        cls.temp_dir = tempfile.TemporaryDirectory()
        cls._build_sqlnet_config()
        cls._build_tnsnames_config()
        oracledb.init_oracle_client(config_dir=cls.temp_dir.name)
        cls._build_user()

    @classmethod
    def tearDownClass(cls):
        cls.temp_dir.cleanup()

    def test_ext_1100(self):
        "E1100 - external authentication with tnsnames alias (implicit)"
        with oracledb.connect(dsn=self.alias_name) as conn:
            with conn.cursor() as cursor:
                cursor.execute("select user from dual")
                (user,) = cursor.fetchone()
                self.assertEqual(user, self.user.upper())

    def test_ext_1101(self):
        "E1101 - external authentication with tnsnames alias (explicit)"
        with oracledb.connect(externalauth=True, dsn=self.alias_name) as conn:
            with conn.cursor() as cursor:
                cursor.execute("select user from dual")
                (user,) = cursor.fetchone()
                self.assertEqual(user, self.user.upper())

    def test_ext_1102(self):
        "E1102 - external authentication with connect string (explicit)"
        with oracledb.connect(
            externalauth=True, dsn=test_env.get_connect_string()
        ) as conn:
            with conn.cursor() as cursor:
                cursor.execute("select user from dual")
                (user,) = cursor.fetchone()
                self.assertEqual(user, self.user.upper())

    def test_ext_1103(self):
        "E1103 - external authentication with tnsnames alias (explicit)"
        pool = oracledb.create_pool(
            externalauth=True, dsn=self.alias_name, homogeneous=False
        )
        with pool.acquire() as conn:
            with conn.cursor() as cursor:
                cursor.execute("select user from dual")
                (user,) = cursor.fetchone()
                self.assertEqual(user, self.user.upper())

    def test_ext_1104(self):
        "E1104 - external authentication with connect string (explicit)"
        pool = oracledb.create_pool(
            externalauth=True,
            homogeneous=False,
            dsn=test_env.get_connect_string(),
        )
        with pool.acquire() as conn:
            with conn.cursor() as cursor:
                cursor.execute("select user from dual")
                (user,) = cursor.fetchone()
                self.assertEqual(user, self.user.upper())


if __name__ == "__main__":
    test_env.run_test_cases()



# -----------------------------------------------------------------------------
# Copyright (c) 2020, 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
3300 - Module for testing Simple Oracle Document Access (SODA) Database
"""

import datetime
import decimal
import json
import unittest

import test_env


@unittest.skipIf(
    test_env.skip_soda_tests(), "unsupported client/server combination"
)
class TestCase(test_env.BaseTestCase):
    def __verify_doc(
        self,
        doc,
        bytes_content,
        str_content=None,
        content=None,
        key=None,
        media_type="application/json",
    ):
        self.assertEqual(doc.getContentAsBytes(), bytes_content)
        if str_content is not None:
            self.assertEqual(doc.getContentAsString(), str_content)
        if content is not None:
            self.assertEqual(doc.getContent(), content)
        self.assertEqual(doc.key, key)
        self.assertEqual(doc.mediaType, media_type)
        self.assertIsNone(doc.version)
        self.assertIsNone(doc.lastModified)
        self.assertIsNone(doc.createdOn)

    def test_3300(self):
        "3300 - test creating documents with JSON data"
        soda_db = self.get_soda_database()
        val = {"testKey1": "testValue1", "testKey2": "testValue2"}
        if test_env.get_client_version() < (23, 4):
            str_val = json.dumps(val)
        else:
            str_val = str(val)
        bytes_val = str_val.encode()
        key = "MyKey"
        media_type = "text/plain"
        doc = soda_db.createDocument(val)
        self.__verify_doc(doc, bytes_val, str_val, val)
        str_val = json.dumps(val)
        bytes_val = str_val.encode()
        doc = soda_db.createDocument(str_val, key)
        self.__verify_doc(doc, bytes_val, str_val, val, key)
        doc = soda_db.createDocument(bytes_val, key, media_type)
        self.__verify_doc(doc, bytes_val, str_val, bytes_val, key, media_type)

    def test_3301(self):
        "3301 - test creating documents with raw data"
        soda_db = self.get_soda_database()
        val = b"<html/>"
        key = "MyRawKey"
        media_type = "text/html"
        doc = soda_db.createDocument(val)
        self.__verify_doc(doc, val)
        doc = soda_db.createDocument(val, key)
        self.__verify_doc(doc, val, key=key)
        doc = soda_db.createDocument(val, key, media_type)
        self.__verify_doc(doc, val, key=key, media_type=media_type)

    def test_3302(self):
        "3302 - test getting collection names from the database"
        soda_db = self.get_soda_database()
        self.assertEqual(soda_db.getCollectionNames(), [])
        names = ["zCol", "dCol", "sCol", "aCol", "gCol"]
        sorted_names = list(sorted(names))
        for name in names:
            soda_db.createCollection(name)
        self.assertEqual(soda_db.getCollectionNames(), sorted_names)
        self.assertEqual(soda_db.getCollectionNames(limit=2), sorted_names[:2])
        self.assertEqual(soda_db.getCollectionNames("a"), sorted_names)
        self.assertEqual(soda_db.getCollectionNames("C"), sorted_names)
        self.assertEqual(
            soda_db.getCollectionNames("b", limit=3), sorted_names[1:4]
        )
        self.assertEqual(soda_db.getCollectionNames("z"), sorted_names[-1:])

    def test_3303(self):
        "3303 - test opening a collection"
        soda_db = self.get_soda_database()
        coll = soda_db.openCollection("CollectionThatDoesNotExist")
        self.assertIsNone(coll)
        created_coll = soda_db.createCollection("TestOpenCollection")
        coll = soda_db.openCollection(created_coll.name)
        self.assertEqual(coll.name, created_coll.name)

    def test_3304(self):
        "3304 - test SodaDatabase repr() and str()"
        soda_db = self.get_soda_database()
        self.assertEqual(
            repr(soda_db), f"<oracledb.SodaDatabase on {self.conn}>"
        )
        self.assertEqual(
            str(soda_db), f"<oracledb.SodaDatabase on {self.conn}>"
        )

    def test_3305(self):
        "3305 - test negative cases for SODA database methods"
        soda_db = self.get_soda_database()
        self.assertRaises(TypeError, soda_db.createCollection)
        self.assertRaises(TypeError, soda_db.createCollection, 1)
        with self.assertRaisesFullCode("ORA-40658"):
            soda_db.createCollection(None)
        with self.assertRaisesFullCode("ORA-40675"):
            soda_db.createCollection("CollMetadata", 7)
        self.assertRaises(TypeError, soda_db.getCollectionNames, 1)

    @unittest.skipIf(
        test_env.get_client_version() < (23, 4), "unsupported data types"
    )
    def test_3306(self):
        "3306 - test creating documents with JSON data using extended types"
        soda_db = self.get_soda_database()
        val = {
            "testKey1": "testValue1",
            "testKey2": decimal.Decimal("12.78"),
            "testKey3": datetime.datetime(2023, 7, 3, 11, 10, 24),
        }
        doc = soda_db.createDocument(val)
        str_val = str(val)
        bytes_val = str_val.encode()
        self.__verify_doc(doc, bytes_val, str_val, val)

    def test_3307(self):
        "3307 - test creating documents with int scalar value"
        soda_db = self.conn.getSodaDatabase()
        val = 144
        str_val = "144"
        bytes_val = b"144"
        key = "MyKey"
        media_type = "application/json"
        doc = soda_db.createDocument(val)
        self.__verify_doc(doc, bytes_val, str_val, val)
        doc = soda_db.createDocument(val, key)
        self.__verify_doc(doc, bytes_val, str_val, val, key)
        doc = soda_db.createDocument(val, key, media_type)
        self.__verify_doc(doc, bytes_val, str_val, val, key, media_type)

    @unittest.skipIf(
        test_env.get_client_version() < (23, 4)
        and test_env.get_server_version() < (23, 4),
        "data types serialized differently",
    )
    def test_3308(self):
        "3308 - test creating documents with float scalar value"
        soda_db = self.conn.getSodaDatabase()
        val = 12.2
        str_val = "12.2"
        bytes_val = b"12.2"
        decimal_val = decimal.Decimal(str_val)
        key = "MyKey"
        media_type = "application/json"
        doc = soda_db.createDocument(val)
        self.__verify_doc(doc, bytes_val, str_val, decimal_val)
        doc = soda_db.createDocument(val, key)
        self.__verify_doc(doc, bytes_val, str_val, decimal_val, key)
        doc = soda_db.createDocument(val, key, media_type)
        self.__verify_doc(
            doc, bytes_val, str_val, decimal_val, key, media_type
        )

    @unittest.skipIf(
        test_env.get_client_version() < (23, 4)
        and test_env.get_server_version() < (23, 4),
        "unsupported data types",
    )
    def test_3309(self):
        "3309 - test creating documents with a list"
        soda_db = self.conn.getSodaDatabase()
        val = [12, "str", b"bytes", [1], {"dict": "3"}]
        decimal_val = [
            decimal.Decimal("12"),
            "str",
            b"bytes",
            [decimal.Decimal("1")],
            {"dict": "3"},
        ]
        str_val = (
            "[Decimal('12'), 'str', b'bytes', [Decimal('1')], {'dict': '3'}]"
        )
        bytes_val = (
            b"[Decimal('12'), 'str', b'bytes', [Decimal('1')], {'dict': '3'}]"
        )
        key = "MyKey"
        media_type = "application/json"
        doc = soda_db.createDocument(val)
        self.__verify_doc(doc, bytes_val, str_val, decimal_val)
        doc = soda_db.createDocument(val, key)
        self.__verify_doc(doc, bytes_val, str_val, decimal_val, key)
        doc = soda_db.createDocument(val, key, media_type)
        self.__verify_doc(
            doc, bytes_val, str_val, decimal_val, key, media_type
        )

    @unittest.skipIf(
        test_env.get_client_version() < (23, 4)
        and test_env.get_server_version() < (23, 4),
        "data types serialized differently",
    )
    def test_3310(self):
        "3310 - test creating documents with a boolean scalar value"
        soda_db = self.conn.getSodaDatabase()
        test_values = [(True, "True", b"True"), (False, "False", b"False")]
        key = "MyKey"
        media_type = "application/json"
        for val, str_val, bytes_val in test_values:
            doc = soda_db.createDocument(val)
            self.__verify_doc(doc, bytes_val, str_val, val)
            doc = soda_db.createDocument(val, key)
            self.__verify_doc(doc, bytes_val, str_val, val, key)
            doc = soda_db.createDocument(val, key, media_type)
            self.__verify_doc(doc, bytes_val, str_val, val, key, media_type)

    @unittest.skipIf(
        test_env.get_client_version() < (23, 4)
        and test_env.get_server_version() < (23, 4),
        "data types serialized differently",
    )
    def test_3311(self):
        "3311 - test creating documents with unsupported types"
        soda_db = self.conn.getSodaDatabase()
        values = [
            tuple([144, 2]),
            set("144"),
            bytearray("omg", "utf-8"),
            complex(2j),
            range(4),
        ]
        for value in values:
            with self.assertRaisesFullCode("DPY-3003"):
                soda_db.createDocument(value)


if __name__ == "__main__":
    test_env.run_test_cases()
